import time
from collections import deque
from threading import Lock

import numpy as np
from PyQt5.QtCore import QRect, pyqtSlot, Qt
from PyQt5.QtGui import QPainter, QImage, QPixmap
from PyQt5.QtWidgets import QOpenGLWidget


class ImageView(QOpenGLWidget):
    scaled = True
    max_fps = 30

    def __init__(self):
        QOpenGLWidget.__init__(self)
        self.paint_lock = Lock()
        self.setAttribute(Qt.WA_AlwaysStackOnTop)
        self.frame_size_rect = QRect(0, 0, self.width(), self.height())
        self.screen_size_rect = QRect(0, 0, self.width(), self.height())
        self.last_processed_frame = QPixmap(1, 1).toImage()
        self.frame_buffer = deque(maxlen=2)
        self.last_frame_time = time.perf_counter_ns()
        self.refresh_timeout = 1e9 / self.max_fps

    @pyqtSlot(np.ndarray)
    def slot_new_frame(self, frame_data):
        if (time.perf_counter_ns() - self.last_frame_time) >= self.refresh_timeout:
            self.last_frame_time = time.perf_counter_ns()
            self.frame_buffer.append(frame_data)
            self.update()

    def resizeGL(self, new_w, new_h):
        frame_w, frame_h = (
            self.last_processed_frame.width(),
            self.last_processed_frame.height(),
        )

        if frame_h > 0 and new_h > 0:
            ri = frame_w / frame_h
            rs = new_w / new_h

            if rs > ri:
                w, h = frame_w * new_h / frame_h, new_h
            else:
                w, h = new_w, frame_h * new_w / frame_w
        else:
            w = new_w
            h = new_h
        self.frame_size_rect = QRect(0, 0, w, h)
        self.screen_size_rect = QRect(0, 0, new_w, new_h)

    @staticmethod
    def frame_to_qimage(frame, color_scheme=QImage.Format_RGB888):
        """
        Convert frame from numpy array format into QImage.

        :param frame: Numpy array
        :type frame: Array
        :return: Converted QImage
        :rtype: QImage
        """
        h, w = frame.shape[:2]
        return QImage(frame.data, w, h, color_scheme)

    def process_next_frame(self):
        """
        Frame processing pipeline.

        Apply filters and transformations for next frame.

        :return: Processed frame as QImage object.
        :rtype: QImage
        """
        frame = self.frame_buffer.pop()
        color_scheme = QImage.Format_RGB888
        if len(frame.shape) == 2:
            color_scheme = QImage.Format_Grayscale8
        return self.frame_to_qimage(frame, color_scheme)

    def paintEvent(self, QPaintEvent):
        # Return If previous paint event is still handled
        if self.paint_lock.locked():
            return
        with self.paint_lock:
            try:
                self.last_processed_frame = self.process_next_frame()
            except IndexError:
                # No new frame to process.
                # We just skip here and use previous processed frame.
                pass
            painter = QPainter(self)
            painter.drawImage(self.frame_size_rect, self.last_processed_frame)



# Responses
RESPONSE_OK = "1"
RESPONSE_ERROR = "0"

# Statuses
STATUS_DISCONNECTED = "Disconnected"
STATUS_CONNECTING = "Connecting"
STATUS_CONNECTED = "Connected"
STATUS_CONFIGURING = "Configuring"
STATUS_CONFIGURED = "Configured"
STATUS_IDLE = "Idle"
STATUS_BUSY = "Busy"
STATUS_ERROR = "Error"
STATUS_DISCONNECTING = "Disconnecting"

READY_DEVICE_STATUSES = (STATUS_CONFIGURED, STATUS_IDLE, STATUS_BUSY)
IDLE_DEVICE_STATUSES = (STATUS_CONFIGURED, STATUS_IDLE)
STATUSES_PRIORITY = (
    STATUS_ERROR,
    STATUS_DISCONNECTING,
    STATUS_DISCONNECTED,
    STATUS_CONNECTING,
    STATUS_CONNECTED,
    STATUS_CONFIGURING,
    STATUS_BUSY,
    STATUS_IDLE,
    STATUS_CONFIGURED,
)

# Macro states
IDLE = 0
BUSY = 1
PAUSED = 2
DONE = 3
STOPPED = 4
ERROR = 5
WARNING = 6

ABORT = "Abort macro"
SKIP = "Skip"

MODE_ALL = 2
MODE_EXTENDED = 1
MODE_SIMPLE = 0

# Attribute values
VALUE = "Value"
TYPE = "Type"
MIN = "Min"
MAX = "Max"
UNIT = "Unit"
DECIMALS = "Decimals"
DESCRIPTION = "Description"
READONLY = "Readonly"
TYPE_LIST = "<type> List"
TYPE_LIST_VALUES = "<type> List values"
TYPE_ARRAY = "<type> Array"
FACTOR = "Factor"
OFFSET = "Offset"
TOLERANCE = "Tolerance"
SETPOINT = "Setpoint"
PREVIOUS = "Previous"
SAVE_CHANGE = "Save"
ALL = "*"
DISPLAY = "display"
IMAGE = "Image"
LOCAL_FILESYSTEM = "Local filesystem"
ALLOW_BROADCAST = "Allow broadcast"

# Basic attributes
ATTR_ID = "Id"
ATTR_DESCRIPTION = "Description"
ATTR_STATUS = "Status"
ATTR_ENABLED = "Enabled"

INSTRUCTION_GET = b"0"
INSTRUCTION_SET = b"1"
INSTRUCTION_COMMAND = b"2"
INSTRUCTION_INIT = b"3"
INSTRUCTION_METHOD = b"4"
INSTRUCTION_POLL = b"5"
INSTRUCTION_PULL = b"6"
INSTRUCTION_PUSH = b"7"
INSTRUCTION_FORWARD = b"8"
INSTRUCTION_SYNC = b"9"

# Tokens
TOKEN_ATTRIBUTE = "A!"
TOKEN_ATTRIBUTE_REPLACE = "R!"
TOKEN_ATTRIBUTE_DELETE = "D!"
TOKEN_ATTRIBUTE_GROUP_DELETE = "GD!"
TOKEN_ALL = "*"

# Words
RISING = "Rising"
FALLING = "Falling"
POSITIVE = "Positive"
NEGATIVE = "Negative"
ROWS = "Rows"
COLUMNS = "Columns"

# Macro steps
MACRO_SET_ATTRIBUTE_STEP = "Set attribute"
MACRO_EXECUTE_METHOD_STEP = "Execute method"

# Config enums
CFG_TEMPLATE = 0

# Zmq message tokens
WORKER_READY = b"1"
MSG_TERMINATE = b"0"
WORKER_DONE = b"2"
MSG_EMPTY = b""
MSG_JSON = b"J"
MSG_ARRAY = b"A"
MSG_BYTES = b"B"
MSG_PICKLE = b"P"
MSG_FILE = b"F"
MSG_ACK = b"ACK"
MSG_FORWARD = b"FRWD"
MSG_ABORT = b"ABT"
MSG_PROGRESS = b"PROG"

OPEN_LOOP = "Open loop"
CLOSE_LOOP = "Close loop"

PROXY = "PROXY"

# Session groups
GROUP_RC = "RocketChat"

# List of general Attributes
ATTR_DETECTOR_DISTANCE = "Detector distance"
ATTR_DISTANCE_SETPOINT = "Detector distance setpoint"
ATTR_DIODE_VOLTAGE = "Diode voltage"
ATTR_FILEWRITER_STATUS = "Filewriter status"
ATTR_EIGERSTREAM_STATUS = "Eigerstream status"
ATTR_COLLIMATOR_STATUS = "Collimator status"
ATTR_POSITION_Z = "Position z"
ATTR_IMAGES_NUMBER = "Images number"
ATTR_EXTERNAL_TRIGGER = "External trigger"
ATTR_ACQUIRED_FRAMES = "Acquired frames"
ATTR_FRAME_NAME = "Frame name"
ATTR_FRAME_FILE = "Frame file"
ATTR_TRIGGERS_SP = "Triggers setpoint"
ATTR_TRIGGERS_GENERATED = "Triggers generated"
ATTR_LAST_TRIGGER_TYPE = "Last trigger type"
TYPE_SW = "SW"
TYPE_HW = "HW"
ATTR_RUN_ID = "Run id"
ATTR_AVAILABLE_LENGTH = "Available length"
ATTR_USED_LENGTH = "Used length"
ATTR_USED_LENGTH_PERCENT = "Used length percentage"
ATTR_REFERENCE_POSITION = "Reference position"
ATTR_TAPE_RADIUS = "Tape radius"
ATTR_SERIAL_NUMBER = "Serial number"
ATTR_PRODUCT_TYPE = "Product type"
ATTR_MODEL = "Model name"
ATTR_ANALOG_INPUT_READOUT_TASK = "Analog readout task"
ATTR_TERMINAL_CONFIGURATION = "Terminal configuration"
ATTR_SAMPLES_NUMBER = "Number of samples"
ATTR_READOUT_AT_SAMPLES = "Readout at samples"
ATTR_TEMPERATURE = "Temperature"
ATTR_TEMPERATURE_SETPOINT = "Temperature setpoint"
ATTR_CHANNELS = "Channels"
ATTR_POSITION = "Position"
ATTR_TARGET_POSITION = "Target position"
ATTR_STALL_FLAG = "Stall flag"
ATTR_MOVING = "Moving"
ATTR_VELOCITY = "Velocity"
ATTR_VELOCITY_SETPOINT = "Velocity setpoint"
ATTR_HOLDING_CURRENT = "Holding current"
ATTR_MAXIMUM_VELOCITY = "Maximum velocity"
ATTR_RUN_CURRENT = "Run current"
ATTR_ROTATION_DIRECTION = "Rotation direction"
ATTR_TORQUE_SPEED = "Torque speed"
ATTR_ENCODER_ENABLED = "Encoder enabled"
ATTR_MICROSTEP_RESOLUTION = "Microstep resolution"
ATTR_STEPS_PER_REVOLUTION = "Steps / revolution"
ATTR_STALL_FACTOR = "Stall factor"
ATTR_POSITIVE_LIMIT_SWITCH = "+limit switch"
ATTR_NEGATIVE_LIMIT_SWITCH = "-limit switch"
ATTR_INDEX_MARK = "Index mark"
ATTR_USE_POSITIVE_LIMIT_SWITCH = "Use +limit switch"
ATTR_USE_NEGATIVE_LIMIT_SWITCH = "Use -limit switch"
ATTR_INITIAL_VELOCITY = "Initial velocity"
ATTR_REFERENCING = "Referencing"
ATTR_ERROR_CODE = "Error code"
ATTR_LATENCY = "Latency"
ATTR_MODULES = "Modules count"
ATTR_STATE = "State"
ATTR_ENCODER_BASE_UNIT = "Enc base unit"
ATTR_HAND_TOOL = "Hand tool"
ATTR_MOVEMENT_LOCKED = "Movement locked"
ATTR_INTERNAL_COMM_FAILURE = "Comm failure"
ATTR_STREAMING = "Streaming"
ATTR_HOLDING = "Holding"
ATTR_SENSOR_PRESENT = "Sensor present"
ATTR_BOOSTER_PRESENT = "Booster present"
ATTR_ADJUSTMENT_ACTIVE = "Adjustment active"
ATTR_IOM_PRESENT = "I/O present"
ATTR_HIGH_VOLTAGE_FAILURE = "High voltage failure"
ATTR_HIGH_VOLTAGE_OVERLOAD = "High voltage overload"
ATTR_OVER_TEMPERATURE = "Over temperature"
ATTR_CLOSED_LOOP_ACTIVE = "Closed loop active"
ATTR_CALIBRATING = "Calibrating"
ATTR_MOVE_DELAYED = "Move delayed"
ATTR_CALIBRATED = "Calibrated"
ATTR_REFERENCED = "Referenced"
ATTR_END_STOP_REACHED = "On end stop"
ATTR_RANGE_LIMIT_REACHED = "On range limit"
ATTR_FOLLOWING_LIMIT_REACHED = "On Following limit"
ATTR_MOVEMENT_FAILED = "Movement failed"
ATTR_POSITIONER_TYPE = "Positioner type"
ATTR_POSITIONER_CODE = "Positioner code"
ATTR_CONFIGURED_LIMITS = "Configured Limits"
ATTR_POSITION_UPPER_LIMIT = "Position +limit"
ATTR_POSITION_LOWER_LIMIT = "Position -limit"
ATTR_STEP_FREQUENCY = "Step frequency"
ATTR_STEP_AMPLITUDE = "Step amplitude"
ATTR_ACCELERATION = "Acceleration"
ATTR_DECELERATION = "Deceleration"
ATTR_SENSOR_MODE = "Sensor mode"
ATTR_MOVE_MODE = "Move mode"
ATTR_SAFE_DIRECTION = "Safe direction"
ATTR_HAND_TOOL_MODE = "Hand tool mode"
ATTR_SENSOR_TYPE = "Sensor type"
ATTR_REVOLUTIONS = "Revolutions"
ATTR_BAUDRATE = "Baudrate"
ATTR_REFERENCING_METHOD = "Referencing method"
ATTR_ACCELERATION_PATTERN = "Acceleration pattern"
ATTR_ACCELERATION_TIME = "Acceleration time"
ATTR_CORRECTION = "Position correction"
ATTR_CORRECTION_RETRY = "Correction retry_count"
ATTR_DECELERATION_TIME = "Deceleration time"
ATTR_ENCODER_ADDING = "Encoder adding"
ATTR_ENCODER_READ = "Encoder read"
ATTR_ENCODER_RESOLUTION = "Encoder resolution"
ATTR_PULSE_POSITION = "Pulse position"
ATTR_PULSE_READ = "Pulse read"
ATTR_START_SPEED = "Start speed"
ATTR_STEP_FACTOR = "Step factor"
ATTR_STEP_RESOLUTION = "Step resolution"
ATTR_TOP_SPEED = "Top speed"
ATTR_TRIGGER_EDGE = "Trigger edge"
ATTR_TRIGGER_ENC_PITCH = "Trigger enc pitch"
ATTR_TRIGGER_LOGIC = "Trigger logic"
ATTR_TRIGGER_PM_PITCH = "Trigger pm pitch"
ATTR_TRIGGER_PULSE_WIDTH = "Trigger pulse width"
ATTR_TRIGGER_SELECT = "Trigger select"
ATTR_TRIGGER_SOURCE = "Trigger source"
ATTR_ENCODER_CALC_NUM = "Encoder calc num"
ATTR_ENCODER_CALC_DEN = "Encoder calc den"
ATTR_ROTATION_CHANGE = "Rotation change"
ATTR_SWAP_LIMITS = "Swap limits"
ATTR_RETRY_COUNT = "Retry count"
ATTR_FEEDBACK_TIMEOUT = "Feedback timeout"
ATTR_FEEDBACK_CONTROL_TYPE = "Feedback type"
ATTR_MOTOR_EXCITATION = "Motor excitation"
ATTR_MOTOR_SELECTION = "Motor selection"
ATTR_MICROSTEP_SELECT = "Microstep select"
ATTR_STOP_TYPE = "Stop type"
ATTR_EMERGENCY_STOP = "Emergency stop"
ATTR_CWL = "CW limit"
ATTR_CCWL = "CCW limit"
ATTR_ORG = "ORG signal"
ATTR_NORG = "NORG signal"
ATTR_LAST_ERROR = "Last error"
ATTR_LOG_DIRECTORY = "Log directory"
ATTR_SCAN_DIRECTORY = "Scan directory"
ATTR_ALLOW_CONSOLE_LOG = "Allow console log"
ATTR_ALLOW_DEVICE_LOG = "Allow device log"
ATTR_ALLOW_GUI_LOG = "Allow gui log"
ATTR_ALLOW_ATTRIBUTE_LOG = "Allow attribute log"
ATTR_ALLOW_MACRO_LOG = "Allow macro log"
ATTR_DEVICES_COUNT = "Devices count"
ATTR_LOOPS_RUNNING = "Control loops running"
ATTR_FILEPATH = "Filepath"
ATTR_CONTENT = "Content"
ATTR_STEP_INDEX = "Step index"
ATTR_POINT_INDEX = "Point index"
ATTR_STEPS_COUNT = "Steps count"
ATTR_POINTS_COUNT = "Points count"
ATTR_PROGRESS = "Progress"
ATTR_STARTED_AT = "Started at"
ATTR_RUNNING_TIME = "Running time"
ATTR_ESTIMATED_TIME = "Estimated time"
ATTR_ESTIMATED_END = "Estimated end"
ATTR_REMAINING_TIME = "Remaining time"
ATTR_CHAIN_LINK = "Chain link"
ATTR_OPENED = "Opened"
ATTR_OPENED_POSITION = "Opened position"
ATTR_CLOSED_POSITION = "Closed position"
ATTR_OPENED_VALUE = "Opened value"
ATTR_CLOSED_VALUE = "Closed value"
ATTR_OPENED_AT = "Opened at"
ATTR_AUTO_MODE = "Auto mode"
ATTR_SYNTAX_VERSION = "Syntax vesrion"
ATTR_SERVO_STATE = "Servo state"
ATTR_ON_TARGET = "On target"
ATTR_ACTIVE_UNIT = "Active unit"
ATTR_REFERENCING_DIRECTION_X = "Ref direction X"
ATTR_REFERENCING_DIRECTION_Y = "Ref direction Y"
ATTR_REFERENCING_DIRECTION_Z = "Ref direction Z"
ATTR_PIVOT_MODE = "Pivot mode"
ATTR_PIVOT_X = "Pivot X"
ATTR_PIVOT_Y = "Pivot Y"
ATTR_PIVOT_Z = "Pivot Z"
ATTR_POSITIONS = "Positions"
ATTR_USE_MARK_INDEX = "Use mark index"
ATTR_BACKLASH_DISTANCE = "Backlash distance"
ATTR_BUFFERED_COMMANDS = "Buffered commands"
ATTR_HANGING_COMMANDS = "Hanging commands"
ATTR_CONSTANT_CURRENT = "Constant current"
ATTR_CONSTANT_VOLTAGE = "Constant voltage"
ATTR_CURRENT = "Current"
ATTR_VOLTAGE = "Voltage"
ATTR_MAX_CURRENT = "Max current"
ATTR_MAX_VOLTAGE = "Max voltage"
ATTR_OUTPUT_SWITCH = "Output switch"
ATTR_INTENSITY = "Intensity"
ATTR_RESOURCE_DIRECTORY = "Resource directory"
ATTR_CACHE_DIRECTORY = "Cache directory"
ATTR_SCAN_PREFIX = "Scan prefix"
ATTR_SCAN_COUNT = "Scan count"
ATTR_OUTPUT_FILENAME = "Output filename"
ATTR_TIMEOUT = "Timeout"
ATTR_ACRONYM = "Acronym"
ATTR_MOTOR_TYPE = "Motor type"
ATTR_AMPLIFIER_GAIN = "Amplifier gain"
ATTR_FAST_REFERENCE_SPEED = "Fast reference speed"
ATTR_FULL_STEP_RESOLUTION = "Full step resolution"
ATTR_HOMING_DIRECTION = "Homing direction"
ATTR_JOG_BACKWARD = "Jog backward"
ATTR_JOG_FORWARD = "Jog forward"
ATTR_LATCH_INPUT_LEVEL = "Latch input level"
ATTR_LIMIT_SWITCH_ACTIVE_LEVEL = "Limit switch active level"
ATTR_LOW_CURRENT_STEPPER_MODE = "Low current stepper mode"
ATTR_SLOW_REFERENCE_SPEED = "Slow reference speed"
ATTR_STEP_DRIVE_RESOLUTION = "Step drive resolution"
ATTR_JOG_MOTOR = "Jog motor"
ATTR_HOMING_METHOD = "Homing method"
ATTR_LIMIT_SWITCHES_MODE = "Limit switches mode"
ATTR_REMOTE_TIMESTAMP = "Remote timestamp"
ATTR_LOCAL_TIMESTAMP = "Local timestamp"
ATTR_DELTA_TIMESTAMP = "Delta timestamp"
ATTR_REMOTE_TRAIN_ID = "Remote train id"
ATTR_LOCAL_TRAIN_ID = "Local train id"
ATTR_MIN_UPDATE_TIMEOUT = "Min update timeout"
ATTR_TRAIN_TIMEOUT = "Train timeout"
ATTR_TRAIN_OFFSET = "Train id offset"
ATTR_COUNTER = "Counter"
ATTR_REFRESHING = "Refreshing"
ATTR_WAVETABLE_LENGTH = "Wavetable length"
ATTR_WAVETABLE_DATA = "Wave table data"
ATTR_WAVETABLE_RATE = "Wave table rate"
ATTR_WAVETABLE_INTERPOLATION = "Wave table interpolation"
ATTR_WAVE_GEN_OFFSET = "Wave gen offset"
ATTR_TRIGGER_OUTPUTS_COUNT = "Trigger outputs count"
ATTR_WAVE_GEN_COUNT = "Wave generators count"
ATTR_WAVE_TABLE = "Wave table"
ATTR_SERVO_UPDATE_TIME = "Servo update time"
ATTR_RECORD_TABLE_RATE = "Record table rate"
ATTR_RECORD_TABLE_DATA = "Record table data"
ATTR_RECORD_TABLE_DATA_SETPOINTS = "Record table data setpoints"
ATTR_WAVETABLE_OFFSET = "Wave table offset"
ATTR_TRIGGER_LINE_DATA = "Trigger line data"
ATTR_SCANNING = "Scanning"
ATTR_READOUT = "Readout"
ATTR_CAMERA_CALIBRATION = "Pixel calibration"
ATTR_COLOR_OUTPUT = "Color output"
ATTR_COMMENT = "Comment"
ATTR_DUMMY_IMAGE = "Dummy image"
ATTR_EXPOSURE_TIME = "Exposure time"
ATTR_FRAMERATE = "Framerate"
ATTR_FRAME_IMAGE = "Frame image"
ATTR_GAIN = "Gain"
ATTR_HEIGHT_DIVIDER = "Height divider"
ATTR_HEIGHT_MAX = "Height max"
ATTR_LIVE_VIEW = "Live view"
ATTR_LIVE_VIEW_FPS_LIMIT = "Live view limit"
ATTR_OUTPUT_NAME = "Output name"
ATTR_PIXEL_BITSIZE = "PixelSize"
ATTR_REAL_FRAMERATE = "Real framerate"
ATTR_ROI_OFFSET_X = "ROI offset X"
ATTR_ROI_OFFSET_Y = "ROI offset Y"
ATTR_SAVE_FRAMES = "Save frames"
ATTR_START_ACQUISITION = "Start acquisition"
ATTR_STOP_ACQUISITION = "Stop acquisition"
ATTR_SUMMED_FRAMES = "Summed frames"
ATTR_TRIGGER_MODE = "Trigger mode"
ATTR_WIDTH_DIVIDER = "Width divider"
ATTR_WIDTH_MAX = "Width max"
GROUP_ROI = "Roi"
ATTR_WAVE_GENERATOR = "Wave generator"
ATTR_OUTPUT_DIRECTORY = "Output directory"
ATTR_TRIGGER_DURATION = "Trigger duration"
ATTR_SETPOINT = "Setpoint"
ATTR_FAULT_BYTE = "Fault byte"
ATTR_FAST_REFERENCE_VELOCITY = "Fast reference velocity"
ATTR_SLOW_REFERENCE_VELOCITY = "Slow reference velocity"
ATTR_REFERENCE_STEPS_BACK = "Reference steps back"
ATTR_HOLDING_TIME = "Holding time"
ATTR_STEPS = "Steps"
ATTR_PRESSURE = "Pressure"
ATTR_POWER = "Power"
ATTR_CURRENT_SAMPLE = "Current sample"
ATTR_CURRENT_CHANNEL = "Current channel"
ATTR_ACTIVE_SWITCHES = "Active switches"
ATTR_LIBRARY = "Library"
ATTR_ENCODER_READOUT = "Encoder readout"
ATTR_COUNTS_PER_REVOLUTION = "Counts pre revolution"
ATTR_REFERENCE_VELOCITY = "Reference velocity"
ATTR_DIRECTION = "Direction"
ATTR_DWELL_TIME = "Dwell time"
ATTR_TEMPLATES_COUNT = "Templates count"
ATTR_MACRO_COUNT = "Macro count"
ATTR_MACRO_PREFIX = "Macro prefix"
ATTR_TEMPLATE_DIRECTORY = "Template directory"
ATTR_UPLOADED_COUNT = "Uploaded count"
ATTR_FINISHED_COUNT = "Finished count"
ATTR_RUNNING_COUNT = "Running count"
ATTR_MAX_RUNNING_MACROS = "Max running macros"
ATTR_MAX_MACRO_HISTORY = "Max macro history"
ATTR_TEMPLATE = "Template"
ATTR_CHAIN_LINK_OUTPUT = "Chain link output"
ATTR_OFF = "Off"
ATTR_PS1_VOLTAGE = "PS 1 voltage"
ATTR_PS1_POWER = "PS 1 power"
ATTR_PS1_CURRENT = "PS 1 current"
ATTR_PS2_VOLTAGE = "PS 2 voltage"
ATTR_PS2_POWER = "PS 2 power"
ATTR_PS2_CURRENT = "PS 2 current"
ATTR_PS3_VOLTAGE = "PS 3 voltage"
ATTR_PS3_POWER = "PS 3 power"
ATTR_PS3_CURRENT = "PS 3 current"
ATTR_PS4_VOLTAGE = "PS 4 voltage"
ATTR_PS4_POWER = "PS 4 power"
ATTR_PS4_CURRENT = "PS 4 current"
ATTR_POWER_SETPOINT = "Power setpoint"
ATTR_VOLTAGE_SETPOINT = "Voltage setpoint"
ATTR_CURRENT_SETPOINT = "Current setpoint"
ATTR_OUTPUT_RELEASED = "Output released"
ATTR_PROCESS_ID = "Process ID"
ATTR_MEMORY_USAGE = "Memory usage"
ATTR_CPU_USAGE = "Cpu usage"
ATTR_AVAILABLE_MEMORY = "Available memory"
ATTR_HOME_SWITCH_FLAG = "Home switch flag"
ATTR_HARDWARE_VERSION = "Hardware version"
ATTR_SOFTWARE_VERSION = "Software version"
ATTR_MANUFACTURE_DATE = "Manufacture date"
ATTR_APPLICATION_NUMBER = "Application number"
ATTR_APPLICATION_VERSION = "Application version"
ATTR_OPERATION_MODE = "Operation mode"
ATTR_VELOCITY_UNITS = "Velocity units"
ATTR_WIDTH = "Width"
ATTR_HEIGHT = "Height"
ATTR_INPUT_DIR_LIST = "Input dir list"
ATTR_OUTPUT_DIR_LIST = "Output dir list"
ATTR_FILE_FILTER_PATTERN = "File filter pattern"
ATTR_ADDRESS = "Address"
ATTR_HOST = "Host"
ATTR_PORT = "Port"
ATTR_TASKS = "Tasks"
ATTR_WORKER = "Worker"
ATTR_CLIENT = "Client"
ATTR_TASK_INSTRUCTION = "Task instruction"
ATTR_SOURCE_FILE = "Source file"
ATTR_TARGET_FILE = "Target file"
ATTR_ABORTED = "Aborted"
ATTR_DRIVES = "Drives"
ATTR_FINISHED_AT = "Finished at"
ATTR_REQUESTED_FILES = "Requested files"
ATTR_SYNCED_FILES = "Synced files"
ATTR_METADATA = "Metadata"
ATTR_LOGFILE = "Logfile"
ATTR_HEADER = "Header"
ATTR_LAST_LOG_LINE = "Last log line"
ATTR_INTERVAL = "Interval"
ATTR_LOGGING = "Logging"
ATTR_MEASURING_SCALE = "Measuring scale"
ATTR_DISTANCE = "Distance"
ATTR_MODES = "Modes"
ATTR_AUTO_STOP = "Auto stop"
ATTR_ABSOLUTE_POSITION = "Absolute position"
ATTR_WARNING = "Warning"
ATTR_VELOCITY_TIMEOUT = "Velocity timeout"
ATTR_SPEED_TABLE_ROW = "Speed table row"
ATTR_RESISTANCE = "Resistance"
ATTR_ALLOW_RC = "Allow RocketChat"
ATTR_MAIN_ROOM = "Main room name"
ATTR_RC_MACRO_DISCUSSION = "Macro ds name"
ATTR_AUTH_TOKEN = "Auth token"
ATTR_USER_NAME = "username"
ATTR_PASSWORD = "Password"

ATTR_ANALOGUE_PROGRAMMING_ACTIVE = "Analogue prog active"
ATTR_CALIBRATION_MODE = "Calibration mode"
ATTR_CONTROL_CURRENT = "Control current"
ATTR_CONTROL_VOLTAGE = "Control voltage"
ATTR_CURRENT_RAMPING_RATE = "Current ramping rate"
ATTR_DIGITAL_PROGRAMMING_ACTIVE = "Digital prog active"
ATTR_MONITORED_CURRENT = "I-Mon"
ATTR_MONITORED_VOLTAGE = "V-Mon"
ATTR_OUTPUT_POLARITY = "Output polarity"
ATTR_OUTPUT_RELEASE = "Output release"
ATTR_POLARITY_REVERSAL = "Polarity"
ATTR_VOLTAGE_RAMPING_RATE = "Voltage ramping rate"
ATTR_VOLTAGE_READOUT = "Voltage readout"
ATTR_CURRENT_READOUT = "Current readout"
ATTR_CURRENT_ON_TARGET = "Current on target"
ATTR_VOLTAGE_ON_TARGET = "Voltage on target"
ATTR_RAMP_VOLTAGE_MODE = "Ramp voltage mode"
ATTR_RAMP_CURRENT_MODE = "Ramp current mode"
ATTR_OUTPUT_POWER = "Output power"
ATTR_DIGITAL_OUTPUT = "Digital output"
ATTR_DIGITAL_INPUT = "Digital input"
ATTR_ANALOG_OUTPUT = "Analog output"
ATTR_ANALOG_INPUT = "Analog input"
ATTR_CHANNELS_CONFIGURATION = "Channels configuration"
ATTR_NAME = "Name"
ATTR_MEASUREMENT_STATUS = "Measurement status"
ATTR_FIRMWARE = "Firmware version"
ATTR_SENSORS_COUNT = "Sensors count"
GROUP_MATERIALS = "Materials"
ATTR_MATERIAL = "Material"
ATTR_AUTO_Z = "Auto Z"
ATTR_OUTPUT = "Output"
GROUP_OUTPUTS = "Outputs"
ATTR_PROCESS = "Process"
GROUP_PROCESSES = "Processes"
GROUP_USER_MESSAGES = "User messages"
ATTR_USER_MESSAGE = "User message"
ATTR_Z_RATIO = "Z ratio"
ATTR_FUNDAMENTAL_FREQUENCY = "Fundamental freq"
ATTR_AHARMONIC_FREQUENCY = "Aharmonic freq"
ATTR_SENSOR_THICKNESS = "Sensor thickness"
ATTR_CRYSTAL_LIFE = "Crystal life"
ATTR_RAW_RATE = "Raw rate"
ATTR_SHUTTER_STROKES = "Shutter strokes"
ATTR_SHUTTER_OPEN = "Shutter open"
ATTR_LAYER_THICKNESS = "Layer thickness"
ATTR_ACTIVE_PROCESS = "Active process"
ATTR_GROUPS = "Groups"
ATTR_LAST_UPDATE = "Last update"
ATTR_SOURCE_ID = "Source id"
GROUP_FRAME = "Frame"
ATTR_FRAME = "Frame"
ATTR_FRAME_COUNT = "Frame count"
ATTR_ALLOW_STREAM = "Allow stream"
ATTR_MAX_STREAM_RATE = "Max stream rate"
ATTR_ALLOW_BROADCAST = "Allow broadcast"
ATTR_MAX_BROADCAST_RATE = "Max broadcast rate"
ATTR_SOURCE = "Source"
ATTR_SOURCE_FRAMERATE = "Source framerate"
ATTR_SOURCE_WIDTH = "Source width"
ATTR_SOURCE_HEIGHT = "Source height"
ATTR_AUTO_SET_FRAMERATE = "Autoset framerate"
ATTR_CAPTURE_API = "Capture API"
ATTR_SOURCES_COUNT = "Sources count"
ATTR_STREAM_READOUT = "Stream readout"
ATTR_VALUE_READOUT = "Value readout"
ATTR_STREAMABLE = "Streamable"
ATTR_FULL_ACCESS = "Full access"
ATTR_PILOT_LASER = "Pilot laser"
ATTR_IS_STABLE = "Is stable"
ATTR_CAPTURE_REGION = "Capture region"
ATTR_FourCC_CODEC = "FourCC Codec"
ATTR_FILENAME_PREFIX = "Filename prefix"
ATTR_RECORDING = "Recording"
ATTR_FRAME_AGGREGATION = "Frame aggregation"
ATTR_BUFFER_AGGREGATION = "Buffer aggregation"
ATTR_STREAMING_MODE = "Streaming mode"
ATTR_BUFFERS_INTERLEAVED = "Buffers interleaved"
ATTR_FILTER_RATE = "Filter rate"
ATTR_BUFFERS_COUNT = "Buffers count"
ATTR_FRAMES_GRABBED = "Frames grabbed"
ATTR_HEAD_FIBER_LENGTH = "Head fiber length"
ATTR_EXTENSION_FIBER_LENGTH = "Extension fiber length"
ATTR_HEAD_TYPE = "Head type"
ATTR_MIN_WORKING_DISTANCE = "Min working distance"
ATTR_MAX_WORKING_DISTANCE = "Max working distance"
ATTR_BEAM_INTERRUPT_TOLERANCE = "Beam interrupt tolerance"
ATTR_CLOSE_LOOP_UNIT = "Close loop unit"
ATTR_CLOSE_LOOP_FACTOR = "Close loop factor"
ATTR_OPEN_LOOP_FACTOR = "Open loop factor"
ATTR_MOVE_METHOD = "Move method"
ATTR_CHANNEL_ENABLED = "Channel enabled"
ATTR_PRESCAN_PIXELS = "Prescan pixels"
ATTR_POSTSCAN_PIXELS = "Postcan pixels"
ATTR_RECORDED_POSITIONS = "Recorded positions"
ATTR_REFERENCE_DIRECTION = "Reference direction"
ATTR_AUTO_CONTROL_SHUTTER = "Auto control shutter"
ATTR_OPEN_DELAY = "Open delay"
ATTR_CLOSE_DELAY = "Close delay"



import time
from random import randint
from threading import Event, Thread

import numpy as np

from kamzik3 import DeviceError
from kamzik3.constants import *
from kamzik3.devices.dummy.deviceDummy import DeviceDummy
from kamzik3.snippets.snippetsDecorators import expose_method
from kamzik3.snippets.snippetsUnits import device_units

ATTR_ACQUIRED_FRAMES = "Acquired frames"
ATTR_FRAME_NAME = "Frame name"
ATTR_FRAME_FILE = "Frame file"


class DummyDetector(DeviceDummy):
    offset = 0

    def __init__(self, device_id=None, config=None):
        self.acquisition_running = Event()
        DeviceDummy.__init__(self, device_id, config)

    def _init_attributes(self):
        DeviceDummy._init_attributes(self)
        self.create_attribute(
            ATTR_EXPOSURE_TIME,
            default_value=0,
            readonly=False,
            default_type=np.float,
            unit="ms",
            min_value=1,
            decimals=1,
        )
        self.create_attribute(
            ATTR_FRAME_IMAGE, default_value=0, readonly=True, default_type=np.uint64
        )
        self.create_attribute(ATTR_FRAME_NAME, default_value="Frame", readonly=False)
        self.create_attribute(ATTR_FRAME_FILE, default_value=0, readonly=True)
        self.create_attribute(
            ATTR_SUMMED_FRAMES,
            default_value=0,
            readonly=False,
            default_type=np.uint64,
            min_value=1,
        )
        self.create_attribute(
            ATTR_ACQUIRED_FRAMES, default_value=0, readonly=True, default_type=np.uint64
        )
        self.create_attribute(
            ATTR_WIDTH, default_value=1024, unit="px", default_type=np.uint64
        )
        self.create_attribute(
            ATTR_HEIGHT, default_value=1024, unit="px", default_type=np.uint64
        )
        npa = np.zeros(dtype=np.uint8, shape=(1024, 1024))
        self.create_attribute(
            "Frame",
            group="Output",
            default_value=npa,
            default_type=TYPE_ARRAY,
            unit="",
            readonly=True,
            display=IMAGE,
            min_broadcast_timeout=100,
        )

    @expose_method()
    def stop(self):
        self.acquisition_running.clear()

    def _generate_frame(self):
        width, height = self.get_value(ATTR_WIDTH), self.get_value(ATTR_HEIGHT)
        frame = np.random.randint(0, 255, dtype=np.uint8, size=(width, height))
        wh, hh = int(width / 2), int(height / 2)
        rw = randint(10, 100)
        frame[:25, :height] = 0
        frame[:width, :25] = 0
        frame[width - 25 :, :height] = 0
        frame[:width, height - 25 :] = 0
        frame[wh - rw : wh + rw, hh - rw : hh + rw] = 0
        return frame

    @expose_method(
        {
            "Name": ATTR_FRAME_NAME,
            "Frames": ATTR_SUMMED_FRAMES,
            "Exposure": ATTR_EXPOSURE_TIME,
        }
    )
    def start_acquisition(self, Name, Frames, Exposure):
        if self.acquisition_running.isSet():
            raise DeviceError("Another acquisition is already running")
        self.acquisition_running.set()
        Frames = device_units(self, ATTR_SUMMED_FRAMES, Frames)
        Exposure = device_units(self, ATTR_EXPOSURE_TIME, Exposure)

        self.set_status(STATUS_BUSY)
        self.set_value(ATTR_FRAME_NAME, Name)
        self.set_value(ATTR_SUMMED_FRAMES, Frames)
        self.set_value(ATTR_EXPOSURE_TIME, Exposure)
        self.set_value(ATTR_ACQUIRED_FRAMES, 0)

        def _acquire():
            exposure_time_increment = 10
            current_exposure_time = 0
            last_frame_number = 0
            while self.acquisition_running.isSet():
                time.sleep(exposure_time_increment * 1e-3)
                current_exposure_time += exposure_time_increment
                acquired_frames = int(current_exposure_time / Exposure.m)
                if last_frame_number != acquired_frames:
                    self.set_value(["Output", "Frame"], self._generate_frame())
                    last_frame_number = acquired_frames

                if acquired_frames >= Frames:
                    self.set_value(ATTR_ACQUIRED_FRAMES, Frames)
                    break
                else:
                    self.set_value(ATTR_ACQUIRED_FRAMES, acquired_frames)

            if self.connected:
                self.acquisition_running.clear()
                self.set_status(STATUS_IDLE)

        Thread(target=_acquire).start()



import time

import numpy as np
import serial

import kamzik3
from kamzik3.constants import *
from kamzik3.devices.devicePort import DevicePort
from kamzik3.devices.deviceSocket import DeviceSocket


class DevicePeakTech(DeviceSocket):
    terminator = b"OK\r"
    push_commands_max = 1

    def _init_attributes(self):
        super()._init_attributes()
        self.create_attribute(ATTR_CONSTANT_CURRENT, readonly=True, default_type=bool)
        self.create_attribute(ATTR_CONSTANT_VOLTAGE, readonly=True, default_type=bool)
        self.create_attribute(
            ATTR_VOLTAGE,
            default_type=np.float16,
            min_value=0,
            decimals=3,
            unit="V",
            set_function=self.set_voltage_output,
        )
        self.create_attribute(
            ATTR_MAX_VOLTAGE,
            default_type=np.float16,
            min_value=0,
            decimals=1,
            unit="V",
            max_value=99.9,
            set_function=self.set_max_voltage_output,
            set_value_when_set_function=False,
        )
        self.create_attribute(
            ATTR_CURRENT,
            default_type=np.float16,
            min_value=0,
            decimals=3,
            unit="A",
            set_function=self.set_current_output,
        )
        self.create_attribute(
            ATTR_MAX_CURRENT,
            default_type=np.float16,
            min_value=0,
            max_value=99.9,
            decimals=1,
            unit="A",
            set_function=self.set_max_current_output,
            set_value_when_set_function=False,
        )
        self.create_attribute(
            ATTR_OUTPUT_SWITCH,
            default_type=bool,
            set_function=self.switch_output,
            set_value_when_set_function=False,
        )

    def handle_configuration(self):
        start_at = time.time()

        def _finish_configuration(*_, **__):
            self._config_commands()
            self._config_attributes()
            self.start_polling()
            self.set_status(STATUS_CONFIGURED)
            self.logger.info(
                "Device configuration took {} sec.".format(time.time() - start_at)
            )

        self.command("GOVP\r")
        self.command("GOCP\r", callback=_finish_configuration)

    def handle_readout(self, readout_buffer):
        """
        Input command: GETD [CR]
        Meaning: The PS Display value is 15V and 16A. It is in CC mode.
        Return value: 150016001[CR] OK[CR]

        Input command: GOVP [CR]
        Meaning: The preset upper limit of output Voltage is 11.1V
        Return value: 111[CR] OK[CR]

        Input command: GOCP [CR]
        Meaning: The preset upper limit of output Current is 11.1A
        Return value: 111[CR] OK[CR]
        """
        command, output, callback, token = super().handle_readout(readout_buffer)

        if command == "GETD\r":
            voltage, current, cv_cc_flag = (
                int(output[:4]) / 100.0,
                int(output[4:8]) / 100,
                int(output[8:]),
            )
            self.set_offsetted_value(ATTR_CONSTANT_VOLTAGE, cv_cc_flag == 0)
            self.set_offsetted_value(ATTR_CONSTANT_CURRENT, cv_cc_flag == 1)
            self.set_offsetted_value(ATTR_VOLTAGE, voltage)
            self.set_offsetted_value(ATTR_CURRENT, current)
            if voltage < 0.75:
                self.set_offsetted_value(ATTR_OUTPUT_SWITCH, False)
            else:
                self.set_offsetted_value(ATTR_OUTPUT_SWITCH, True)
        elif command == "GOVP\r":
            max_voltage = int(output[:3]) / 10
            self.set_attribute((ATTR_VOLTAGE, MAX), max_voltage)
            self.set_offsetted_value(ATTR_MAX_VOLTAGE, max_voltage)
        elif command == "GOCP\r":
            max_current = int(output[:3]) / 10
            self.set_attribute((ATTR_CURRENT, MAX), max_current)
            self.set_offsetted_value(ATTR_MAX_CURRENT, max_current)

        if callback is not None:
            self.handle_readout_callback(callback, command, readout_buffer)
        if token:
            kamzik3.session.publisher.push_message(
                self.device_id, [command, readout_buffer], token
            )

    def start_polling(self):
        super().start_polling()
        self.poll_command("GETD\r", 200)

    def set_max_voltage_output(self, value, callback=None):
        """
        Preset the upper limit of output Voltage.

        <voltage>=000<???<Max Volt where Max Volt value refer to product specification.

        Input command: SOVP 151[CR]
        Meaning: Preset upper limit of output Voltage as 15.1V
        Return value: OK[CR]
        """
        self.logger.info("Set max voltage output to {} V".format(value))
        self.command(
            "SOVP{0:03d}\r".format(int(float(value) * 10.0)), callback, with_token=True
        )
        return self.command("GOVP\r")

    def set_max_current_output(self, value, callback=None):
        """
        Preset the upper limit of output Current.

        <current>=000<???<Max Curr where Max Curr value refer to product specification.

        Input command: SOCP 151[CR]
        Meaning: Preset upper limit of output Current as 15.1A
        Return value: OK[CR]
        """
        self.logger.info("Set max current output to {} A".format(value))
        self.command(
            "SOCP{0:03d}\r".format(int(float(value) * 10.0)), callback, with_token=True
        )
        return self.command("GOCP\r")

    def set_voltage_output(self, value, callback=None):
        """
        Preset Voltage value.

        <voltage>=000<???<Max Volt where Max Volt value refer to product specification.
        Input command: VOLT 127[CR]
        Meaning: Set Voltage value as 12.7V
        Return value: OK[CR]
        """
        self.logger.info("Set voltage output to {} V".format(value))
        return self.command(
            "VOLT{0:03d}\r".format(int(float(value) * 10.0)), callback, with_token=True
        )

    def set_current_output(self, value, callback=None):
        """
        Preset Current value.

        <current>=000<???<Max Curr where Max Curr value refer to product specification.

        Input command: CURR 120[CR]
        Meaning: Set Current value as 12.0A
        Return value: OK[CR]
        """
        self.logger.info("Set current output to {} A".format(value))
        return self.command(
            "CURR{0:03d}\r".format(int(float(value) * 10.0)), callback, with_token=True
        )

    def switch_output(self, value, callback=None):
        """
        Switch on/off the output of PS.

        <status>=0/1 (0=ON, 1=OFF)

        Input command: SOUT0 [CR]
        Meaning: Switch on the output of PS
        Return value: OK[CR]
        """
        self.logger.info("Switch output to {}".format(value))
        return self.command(
            "SOUT{}\r".format(int(not value)), callback, with_token=True
        )

    def stop(self):
        self.set_voltage_output(0)
        self.set_current_output(0)


# abstract methods `collect_incoming_data` and `found_terminator` are defined in
# DeviceSocket (base class of DevicePeakTech). Since multiple inheritance is used here
# we better disable the pylint warning.
# pylint: disable=abstract-method
# see https://github.com/python/mypy/issues/9319 for mypy error
class DevicePeakTechPort(DevicePort, DevicePeakTech):  # type: ignore
    terminator = DevicePeakTech.terminator
    push_commands_max = DevicePeakTech.push_commands_max

    def __init__(
        self,
        port,
        baud_rate=9600,
        parity=serial.PARITY_NONE,
        stop_bits=serial.STOPBITS_ONE,
        byte_size=serial.EIGHTBITS,
        device_id=None,
        config=None,
    ):
        super().__init__(
            port, baud_rate, parity, stop_bits, byte_size, device_id, config
        )

    def handshake(self):
        self.logger.info("Handshake initiated")
        self.push(b"GETD\r")
        time.sleep(0.1)  # Give some time for devices to reply
        if self.serial_port.in_waiting != 0:
            readout = self.read_all().split("\r")
            if readout[1] == "OK":
                return True
        return False



import time
from _ctypes import byref
from ctypes import (
    Array,
    create_string_buffer,
    c_uint32,
    CDLL,
    c_uint16,
    c_int,
    c_int8,
    c_uint8,
    c_int64,
)
from typing import Optional

import bidict
import numpy as np
from pint import UndefinedUnitError

from kamzik3.snippets.snippetsUnits import device_units

from kamzik3 import DeviceError, DeviceUnitError
from kamzik3.constants import *
from kamzik3.devices.device import Device
from kamzik3.devices.deviceChannel import DeviceChannel
from kamzik3.devices.deviceOsLibrary import DeviceOsLibrary
from kamzik3.snippets.snippetsDecorators import expose_method
from kamzik3.snippets.snippetsTimer import PreciseCallbackTimer

MODE_PROFILE_POSITION = 1
MODE_PROFILE_VELOCITY = 3
MODE_HOMING = 6
MODE_POSITION = 8
MODE_VELOCITY = 9
MODE_CURRENT = 10

GROUP_VELOCITY_PROFILE = "Velocity profile"
GROUP_POSITION_PROFILE = "Position profile"

_DEVICE_DISCONNECTED = 268435459


class DeviceEpos(Device):
    def __init__(
        self,
        library,
        device_name,
        protocol,
        interface,
        port,
        device_id=None,
        config=None,
    ):
        self.device_name = (
            create_string_buffer(device_name.encode())
            if not isinstance(device_name, Array)
            else device_name
        )
        self.protocol = (
            create_string_buffer(protocol.encode())
            if not isinstance(protocol, Array)
            else protocol
        )
        self.interface = (
            create_string_buffer(interface.encode())
            if not isinstance(interface, Array)
            else interface
        )
        self.port = (
            create_string_buffer(port.encode()) if not isinstance(port, Array) else port
        )
        self.handle = 0
        self.pError = c_uint32()

        if not hasattr(self, "os_library"):
            self.os_library = library
        if not hasattr(self, "library"):
            self.library = None
        super().__init__(device_id, config)
        self.connect()

    def _init_attributes(self):
        Device._init_attributes(self)
        self.create_attribute(
            ATTR_LIBRARY, default_value=self.os_library, readonly=True
        )

    def handle_configuration(self):
        start_at = time.time()

        def _finish_configuration(*_, **__):
            self._config_commands()
            self._config_attributes()
            self.start_polling()
            self.set_status(STATUS_IDLE)
            self.logger.info(
                "Device configuration took {} sec.".format(time.time() - start_at)
            )

        self.connected = True
        _finish_configuration()

    def connect(self, *args):
        """
        Call only this function to connect devices to port / socket / library / ...

        :param args: connect attributes
        """
        self.connecting = True
        try:
            self.library = CDLL(self.os_library)
            self.handle = self.check_error(
                self.library.VCS_OpenDevice(
                    byref(self.device_name),
                    byref(self.protocol),
                    byref(self.interface),
                    byref(self.port),
                    byref(self.pError),
                )
            )
        except OSError as e:
            raise DeviceError("Error loading library {}. {}".format(self.os_library, e))

        self.handle_configuration_event()
        self.connected = True
        self.connecting = False
        self.set_status(STATUS_IDLE)

    def check_error(self, error_code, raise_exception=True):
        """
        Check error response for executed command.

        :param error_code: error code response
        :param raise_exception: raise exception if True
        :return:
        """
        if error_code == 0:
            error_message, error_code = self.get_error()
            if raise_exception:
                if error_code != 0 and error_code != _DEVICE_DISCONNECTED:
                    raise DeviceError("Error: %s" % error_message)
        return error_code

    def get_error(self, pError=None):
        """
        Get error description of given code.

        :param pError: error code
        :return:
        """
        if pError is None:
            pError = self.pError
        pErrorInfo = create_string_buffer(255)
        if self.library.VCS_GetErrorInfo(pError, byref(pErrorInfo), 255) != 0:
            return "Code: %s, %s" % (pError.value, pErrorInfo.value), pError.value
        else:
            return "Code: %s, %s" % (pError.value, "Unknown error"), pError.value

    def disconnect(self):
        self.library.VCS_CloseDevice(self.handle, byref(self.pError))
        return super().disconnect()


class DeviceEposLinux(DeviceEpos):
    def __init__(
        self, device_name, protocol, interface, port, device_id=None, config=None
    ):
        DeviceEpos.__init__(
            self,
            "libEposCmd.so",
            device_name,
            protocol,
            interface,
            port,
            device_id,
            config,
        )


class DeviceEposChannel(DeviceChannel):
    polling = False
    operation_modes = bidict.frozenbidict(
        {
            MODE_PROFILE_POSITION: "Profile position",
            MODE_PROFILE_VELOCITY: "Profile velocity",
            MODE_HOMING: "Homing",
            MODE_POSITION: "Position",
            MODE_VELOCITY: "Velocity",
            MODE_CURRENT: "Current",
        }
    )
    velocity_units = bidict.frozenbidict(
        {
            "Standard": 0,
            "Deci": -1,
            "Centi": -2,
            "Milli": -3,
        }
    )

    def __init__(self, device, channel, device_id=None, config=None) -> None:
        super().__init__(device, channel, device_id, config)
        self.polling_timer: Optional[PreciseCallbackTimer] = None

    def _init_attributes(self):
        DeviceChannel._init_attributes(self)

        self.create_attribute(
            ATTR_HARDWARE_VERSION,
            default_value=None,
            readonly=True,
            default_type=np.uint16,
        )
        self.create_attribute(
            ATTR_SOFTWARE_VERSION,
            default_value=None,
            readonly=True,
            default_type=np.uint16,
        )
        self.create_attribute(
            ATTR_APPLICATION_NUMBER,
            default_value=None,
            readonly=True,
            default_type=np.uint16,
        )
        self.create_attribute(
            ATTR_APPLICATION_VERSION,
            default_value=None,
            readonly=True,
            default_type=np.uint32,
        )
        self.create_attribute(
            ATTR_AUTO_STOP, default_value=True, readonly=False, default_type=bool
        )
        self.create_attribute(
            ATTR_ON_TARGET, default_value=True, readonly=True, default_type=bool
        )
        self.create_attribute(ATTR_REVOLUTIONS, readonly=True, default_type=np.int64)
        self.create_attribute(
            ATTR_OPERATION_MODE,
            default_value=None,
            readonly=True,
            default_type=self.operation_modes.inverse.keys(),
            set_function=self.set_operation_mode,
        )
        self.create_attribute(
            ATTR_MOTOR_EXCITATION,
            default_value=False,
            default_type=bool,
            set_function=self.set_motor_enabled,
        )
        self.create_attribute(
            ATTR_VELOCITY_UNITS,
            default_value=None,
            readonly=True,
            default_type=self.velocity_units.keys(),
            set_function=self.set_velocity_units,
        )
        self.create_attribute(
            ATTR_VELOCITY,
            default_value=None,
            readonly=False,
            unit="rpm",
            default_type=float,
            set_function=self.set_velocity,
        )
        self.create_attribute(
            ATTR_ABSOLUTE_POSITION,
            default_value=None,
            readonly=False,
            unit="deg",
            default_type=float,
            set_function=self.set_absolute_position,
        )
        self.create_attribute(
            ATTR_POSITION,
            default_value=None,
            readonly=False,
            unit="deg",
            default_type=float,
            set_function=self.set_position,
        )
        self.create_attribute(
            ATTR_ACCELERATION,
            default_value=None,
            readonly=False,
            group=GROUP_VELOCITY_PROFILE,
            default_type=float,
            set_function=lambda v: self.set_velocity_profile(acceleration=v),
        )
        self.create_attribute(
            ATTR_DECELERATION,
            default_value=None,
            readonly=False,
            group=GROUP_VELOCITY_PROFILE,
            default_type=float,
            set_function=lambda v: self.set_velocity_profile(deceleration=v),
        )
        self.create_attribute(
            ATTR_ACCELERATION,
            default_value=None,
            readonly=False,
            group=GROUP_POSITION_PROFILE,
            default_type=float,
            set_function=lambda v: self.set_position_profile(acceleration=v),
        )
        self.create_attribute(
            ATTR_DECELERATION,
            default_value=None,
            readonly=False,
            group=GROUP_POSITION_PROFILE,
            default_type=float,
            set_function=lambda v: self.set_position_profile(deceleration=v),
        )
        self.create_attribute(
            ATTR_VELOCITY,
            default_value=None,
            readonly=False,
            group=GROUP_POSITION_PROFILE,
            default_type=float,
            unit="rpm",
            decimals=4,
            set_function=lambda v: self.set_position_profile(velocity=v),
        )

    def handle_configuration(self):
        if self.configured:
            return

        start_at = time.time()

        def _finish_configuration(*_, **__):
            self._config_commands()
            self._config_attributes()
            self.configured = True
            self.start_polling()
            self.set_status(STATUS_CONFIGURED)
            self.logger.info(
                "Device configuration took {} sec.".format(time.time() - start_at)
            )

        self.get_device_firmware_version()
        self.get_operation_mode()
        self.get_enabled_state()
        self.get_velocity_units()
        self.get_velocity_profile()
        self.get_position_profile()
        _finish_configuration()

    def get_device_firmware_version(self):
        """
        Get the device firmware version.

        VCS_GetVersion(
        HANDLE KeyHandle,
        WORD NodeId,
        WORD * pHardwareVersion,
        WORD * pSoftwareVersion,
        WORD * pApplicationNumber,
        WORD * pApplicationVersion,
        DWORD * pErrorCode
        )
        """
        hw_version = c_uint16()
        sw_version = c_uint16()
        app_number = c_uint16()
        app_version = c_uint32()
        error_code = self.device.check_error(
            self.device.library.VCS_GetVersion(
                self.device.handle,
                self.channel,
                byref(hw_version),
                byref(sw_version),
                byref(app_number),
                byref(app_version),
                byref(self.device.pError),
            )
        )

        if error_code == 0 or error_code == _DEVICE_DISCONNECTED:
            # EPOS is disconnected
            return (
                self.get_attribute(ATTR_HARDWARE_VERSION)[VALUE],
                self.get_attribute(ATTR_SOFTWARE_VERSION)[VALUE],
                self.get_attribute(ATTR_APPLICATION_NUMBER)[VALUE],
                self.get_attribute(ATTR_APPLICATION_VERSION)[VALUE],
            )

        self.set_value(ATTR_HARDWARE_VERSION, hw_version.value)
        self.set_value(ATTR_SOFTWARE_VERSION, sw_version.value)
        self.set_value(ATTR_APPLICATION_NUMBER, app_number.value)
        self.set_value(ATTR_APPLICATION_VERSION, app_version.value)

    def get_movement_state(self):
        state = c_int()
        error_code = self.device.check_error(
            self.device.library.VCS_GetMovementState(
                self.device.handle,
                self.channel,
                byref(state),
                byref(self.device.pError),
            )
        )

        if error_code == 0 or error_code == _DEVICE_DISCONNECTED:
            # EPOS is disconnected
            return self.get_attribute(ATTR_ON_TARGET)[VALUE]

        self.set_value(ATTR_ON_TARGET, bool(state.value))
        return state.value

    def get_state(self):
        state = c_uint32()
        error_code = self.device.check_error(
            self.device.library.VCS_GetState(
                self.device.handle,
                self.channel,
                byref(state),
                byref(self.device.pError),
            )
        )

        if error_code == 0 or error_code == _DEVICE_DISCONNECTED:
            # EPOS is disconnected
            self.set_status(STATUS_DISCONNECTED)
            return state.value

        self.set_status(STATUS_BUSY if state.value else STATUS_IDLE)
        return state.value

    def get_enabled_state(self):
        state = c_int8()
        error_code = self.device.check_error(
            self.device.library.VCS_GetEnableState(
                self.device.handle,
                self.channel,
                byref(state),
                byref(self.device.pError),
            )
        )

        if error_code == 0 or error_code == _DEVICE_DISCONNECTED:
            # EPOS is disconnected
            return self.get_attribute(ATTR_MOTOR_EXCITATION)[VALUE]

        self.set_offsetted_value(ATTR_MOTOR_EXCITATION, state.value)
        return state.value

    def get_operation_mode(self):
        operation_mode = c_uint32()
        error_code = self.device.check_error(
            self.device.library.VCS_GetOperationMode(
                self.device.handle,
                self.channel,
                byref(operation_mode),
                byref(self.device.pError),
            )
        )

        if error_code == 0 or error_code == _DEVICE_DISCONNECTED:
            # EPOS is disconnected
            return self.get_attribute(ATTR_OPERATION_MODE)[VALUE]

        self.set_offsetted_value(
            ATTR_OPERATION_MODE, self.operation_modes.get(operation_mode.value, None)
        )
        return operation_mode.value

    def get_velocity_units(self):
        dimension = c_uint8()
        notation = c_int8()
        error_code = self.device.check_error(
            self.device.library.VCS_GetVelocityUnits(
                self.device.handle,
                self.channel,
                byref(dimension),
                byref(notation),
                byref(self.device.pError),
            )
        )

        if error_code == 0 or error_code == _DEVICE_DISCONNECTED:
            # EPOS is disconnected
            return self.get_attribute(ATTR_VELOCITY_UNITS)[VALUE]

        self.set_offsetted_value(
            ATTR_VELOCITY_UNITS, self.velocity_units.inverse.get(notation.value)
        )
        return self.velocity_units.inverse.get(notation.value)

    def get_velocity(self):
        velocity = c_int64()
        error_code = self.device.check_error(
            self.device.library.VCS_GetVelocityIsAveraged(
                self.device.handle,
                self.channel,
                byref(velocity),
                byref(self.device.pError),
            )
        )

        if error_code == 0 or error_code == _DEVICE_DISCONNECTED:
            # EPOS is disconnected
            return self.get_attribute(ATTR_VELOCITY)[VALUE]

        self.set_offsetted_value(ATTR_VELOCITY, velocity.value)
        return velocity.value

    def get_velocity_profile(self):
        acceleration = c_int()
        deceleration = c_int()
        error_code = self.device.check_error(
            self.device.library.VCS_GetVelocityProfile(
                self.device.handle,
                self.channel,
                byref(acceleration),
                byref(deceleration),
                byref(self.device.pError),
            )
        )

        if error_code == 0 or error_code == _DEVICE_DISCONNECTED:
            # EPOS is disconnected
            return (
                self.get_attribute(ATTR_ACCELERATION)[VALUE],
                self.get_attribute(ATTR_DECELERATION)[VALUE],
            )

        self.set_offsetted_value(
            [GROUP_VELOCITY_PROFILE, ATTR_ACCELERATION], acceleration.value
        )
        self.set_offsetted_value(
            [GROUP_VELOCITY_PROFILE, ATTR_DECELERATION], deceleration.value
        )
        return acceleration.value, deceleration.value

    def get_position(self):
        position = c_int64()
        error_code = self.device.check_error(
            self.device.library.VCS_GetPositionIs(
                self.device.handle,
                self.channel,
                byref(position),
                byref(self.device.pError),
            )
        )
        attribute = self.get_attribute(ATTR_ABSOLUTE_POSITION)

        if error_code == 0 or error_code == _DEVICE_DISCONNECTED:
            # EPOS is disconnected
            return attribute[VALUE]

        self.set_offsetted_value(ATTR_ABSOLUTE_POSITION, position.value)
        self.set_offsetted_value(
            ATTR_POSITION, attribute.apply_offset_factor(position.value) % 360
        )
        self.set_offsetted_value(
            ATTR_REVOLUTIONS,
            int(attribute.apply_offset_factor(position.value) / 360),
        )
        return position.value

    def get_position_profile(self):
        velocity = c_int()
        acceleration = c_int()
        deceleration = c_int()
        error_code = self.device.check_error(
            self.device.library.VCS_GetPositionProfile(
                self.device.handle,
                self.channel,
                byref(velocity),
                byref(acceleration),
                byref(deceleration),
                byref(self.device.pError),
            )
        )

        if error_code == 0 or error_code == _DEVICE_DISCONNECTED:
            # EPOS is disconnected
            return (
                self.get_attribute(ATTR_VELOCITY)[VALUE],
                self.get_attribute(ATTR_ACCELERATION)[VALUE],
                self.get_attribute(ATTR_DECELERATION)[VALUE],
            )

        self.set_offsetted_value(
            [GROUP_POSITION_PROFILE, ATTR_VELOCITY], velocity.value
        )
        self.set_offsetted_value(
            [GROUP_POSITION_PROFILE, ATTR_ACCELERATION], acceleration.value
        )
        self.set_offsetted_value(
            [GROUP_POSITION_PROFILE, ATTR_DECELERATION], deceleration.value
        )
        return velocity.value, acceleration.value, deceleration.value

    def set_operation_mode(self, value):
        operation_mode = self.operation_modes.inverse[value]
        self.device.check_error(
            self.device.library.VCS_SetOperationMode(
                self.device.handle,
                self.channel,
                c_int8(operation_mode),
                byref(self.device.pError),
            )
        )
        self.get_operation_mode()

    def set_velocity(self, velocity):
        velocity = int(velocity)
        self.logger.debug("Set velocity to: {} RPM".format(velocity))
        self.set_motor_enabled(True)
        self.set_operation_mode(self.operation_modes[MODE_PROFILE_VELOCITY])
        self.device.check_error(
            self.device.library.VCS_MoveWithVelocity(
                self.device.handle,
                self.channel,
                c_int64(velocity),
                byref(self.device.pError),
            )
        )

    def set_absolute_position(self, position, absolute=True):
        position = int(position)
        self.logger.debug("Set absolute position to: {} deg".format(position))
        self.set_motor_enabled(True)
        self.set_operation_mode(self.operation_modes[MODE_PROFILE_POSITION])
        self.device.check_error(
            self.device.library.VCS_MoveToPosition(
                self.device.handle,
                self.channel,
                c_int64(position),
                c_int(absolute),
                c_int(True),
                byref(self.device.pError),
            )
        )

    def set_position(self, position):
        position = float(position)
        self.logger.debug("Set position to: {} deg".format(position))
        attribute = self.get_attribute(ATTR_ABSOLUTE_POSITION)
        self.set_absolute_position(
            attribute.remove_offset_factor(
                (self[ATTR_REVOLUTIONS].value().m * 360) + position
            )
        )

    def set_motor_enabled(self, value):
        """
        Enable motor for remote control.

        :param value: enable flag (0 - Off, 1 - On)
        :type value: int
        """
        if value:
            self.logger.info("Motor enabled")
            self.device.check_error(
                self.device.library.VCS_SetEnableState(
                    self.device.handle, self.channel, byref(self.device.pError)
                )
            )
        else:
            self.logger.info("Motor disabled")
            self.device.check_error(
                self.device.library.VCS_SetDisableState(
                    self.device.handle, self.channel, byref(self.device.pError)
                )
            )

    def set_velocity_units(self, value):
        self.set_motor_enabled(False)
        dimension = c_uint8()
        notation = c_int8()
        self.device.check_error(
            self.device.library.VCS_GetVelocityUnits(
                self.device.handle,
                self.channel,
                byref(dimension),
                byref(notation),
                byref(self.device.pError),
            )
        )
        notation = c_int8(self.velocity_units.get(value, 0))
        self.device.check_error(
            self.device.library.VCS_SetVelocityUnits(
                self.device.handle,
                self.channel,
                dimension,
                notation,
                byref(self.device.pError),
            )
        )

    def set_velocity_profile(self, acceleration=None, deceleration=None):
        acceleration = c_int(
            int(
                self.get_value([GROUP_VELOCITY_PROFILE, ATTR_ACCELERATION])
                if acceleration is None
                else acceleration
            )
        )
        deceleration = c_int(
            int(
                self.get_value([GROUP_VELOCITY_PROFILE, ATTR_DECELERATION])
                if deceleration is None
                else deceleration
            )
        )
        self.device.check_error(
            self.device.library.VCS_SetVelocityProfile(
                self.device.handle,
                self.channel,
                acceleration,
                deceleration,
                byref(self.device.pError),
            )
        )

    def set_position_profile(self, velocity=None, acceleration=None, deceleration=None):
        velocity = c_int(
            int(
                self.get_value([GROUP_VELOCITY_PROFILE, ATTR_VELOCITY])
                if velocity is None
                else velocity
            )
        )
        acceleration = c_int(
            int(
                self.get_value([GROUP_VELOCITY_PROFILE, ATTR_ACCELERATION])
                if acceleration is None
                else acceleration
            )
        )
        deceleration = c_int(
            int(
                self.get_value([GROUP_VELOCITY_PROFILE, ATTR_DECELERATION])
                if deceleration is None
                else deceleration
            )
        )
        self.device.check_error(
            self.device.library.VCS_SetPositionProfile(
                self.device.handle,
                self.channel,
                velocity,
                acceleration,
                deceleration,
                byref(self.device.pError),
            )
        )

    @expose_method()
    def stop(self):
        self.device.check_error(
            self.device.library.VCS_HaltVelocityMovement(
                self.device.handle, self.channel, byref(self.device.pError)
            )
        )
        self.set_motor_enabled(False)

    def poll_callback(self):
        self.get_velocity()
        self.get_position()
        self.get_enabled_state()
        self.get_state()
        if (
            self.get_value(ATTR_AUTO_STOP)
            and self.get_value(ATTR_OPERATION_MODE)
            == self.operation_modes[MODE_PROFILE_POSITION]
            and self.get_movement_state()
        ):
            self.stop()
        elif (
            self.get_value(ATTR_AUTO_STOP)
            and self.get_value(ATTR_OPERATION_MODE)
            == self.operation_modes[MODE_PROFILE_VELOCITY]
            and self.get_movement_state()
            and self.get_value(ATTR_VELOCITY) == 0
            and self.get_attribute([ATTR_VELOCITY, SETPOINT]) == 0
        ):
            self.stop()

    def start_polling(self):
        self.polling = True
        self.polling_timer = PreciseCallbackTimer(300, self.poll_callback)
        self.polling_timer.start()

    def stop_polling(self):
        self.polling = False
        if self.polling_timer is not None:
            self.polling_timer.stop()

    @expose_method({"position": ATTR_POSITION})
    def move_absolute(self, position):
        try:
            position = device_units(self, ATTR_POSITION, position)
            self.logger.info(
                "Move motor channel to absolute position {}".format(position)
            )
            return self.set_value(ATTR_POSITION, position.m)
        except (UndefinedUnitError, ValueError) as e:
            self.logger.exception(e)
            raise DeviceUnitError(e)

    @expose_method({"position": ATTR_POSITION})
    def move_relative(self, position):
        try:
            position = device_units(self, ATTR_POSITION, position)
            self.logger.info("Move by relative position {}".format(position))
            return self.set_value(ATTR_POSITION, self[ATTR_POSITION].value() + position)
        except (UndefinedUnitError, ValueError) as e:
            self.logger.exception(e)
            raise DeviceUnitError(e)

    @expose_method()
    def reset_position(self):
        self.set_operation_mode(self.operation_modes[MODE_HOMING])
        self.device.check_error(
            self.device.library.VCS_FindHome(
                self.device.handle, self.channel, c_uint8(35), byref(self.device.pError)
            )
        )



import time

import cv2

from kamzik3.devices.general.deviceGrabber import DeviceGrabber

try:
    import mss
except ModuleNotFoundError:
    raise Exception("Module mss was not found.")
import numpy as np

# pylint: disable=ungrouped-imports
from kamzik3.constants import *


class DeviceScreenGrabber(DeviceGrabber):
    def __init__(self, device_id=None, config=None):
        self.region = {"top": 0, "left": 0, "width": 640, "height": 480}
        DeviceGrabber.__init__(self, device_id, config)

    def _init_attributes(self):
        DeviceGrabber._init_attributes(self)
        self.create_attribute(
            ATTR_CAPTURE_REGION,
            default_value="0, 0, 640, 480",
            unit="px",
            set_function=self.set_region,
        )

    def set_region(self, value):
        top, left, width, height = value.split(",")
        # Part of the screen to capture
        self.region = {
            "top": int(top),
            "left": int(left),
            "width": int(width),
            "height": int(height),
        }

    def _acquire_loop(self):
        self.stopped.clear()
        self.frame_delay = 1.0 / self.get_value(ATTR_FRAMERATE)
        self.set_status(STATUS_BUSY)
        time_reference = time.perf_counter_ns()
        with mss.mss() as sct:
            while not self.stopped.wait(
                self.frame_delay - (time.perf_counter_ns() - time_reference) * 1e-9
            ):
                img = np.array(sct.grab(self.region))
                # Don't know why pylint can't find cvtColor and COLOR_BGR2RGB
                # pylint: disable=no-member
                self.set_value(
                    ["Output", "Frame"], cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
                )
                time_reference += self.frame_delay * 1e9



import time
from threading import Event, Thread
import numpy as np
from kamzik3.constants import *
from kamzik3.devices.device import Device
from kamzik3.snippets.snippetsDecorators import expose_method


class DeviceGrabber(Device):
    """Here comes some description of the class DeviceGrabber."""

    def __init__(self, device_id=None, config=None):
        self.stopped = Event()
        self.frame_delay = 0
        Device.__init__(self, device_id, config)
        self.connect()

    def _init_attributes(self):
        Device._init_attributes(self)
        npa = np.zeros(dtype=np.uint8, shape=(1, 1, 3))
        self.create_attribute(
            ATTR_FRAME,
            group="Grabber",
            default_value=npa,
            default_type=TYPE_ARRAY,
            unit="",
            readonly=True,
            display=IMAGE,
            min_broadcast_timeout=100,
        )
        self.create_attribute(
            ATTR_MAX_BROADCAST_RATE,
            default_value=10,
            default_type=np.uint8,
            unit="Hz",
            min_value=0,
            max_value=10000,
            set_function=self.set_max_broadcast_rate,
        )
        self.create_attribute(
            ATTR_FRAMERATE,
            default_value=10,
            default_type=np.float,
            unit="FPS",
            min_value=1,
            max_value=10000,
            set_function=self.set_framerate,
            decimals=2,
        )
        self.create_attribute(
            ATTR_FRAMES_GRABBED, default_value=0, default_type=np.uint, readonly=True
        )

    @expose_method()
    def acquire(self):
        if self.get_value(ATTR_STATUS) == STATUS_BUSY:
            return
        Thread(target=self._acquire_loop).start()

    def _acquire_loop(self):
        """
        Here comes a short description about _acquire_loop.

        self.stopped.clear()
        self.frame_delay = 1. / self.get_value(ATTR_FRAMERATE)
        self.set_status(STATUS_BUSY)

        fps_counter_reference = time_reference = time.perf_counter_ns()
        grabbed_frames_count = 0
        fps_counter = 0
        while not self.stopped.wait((self.frame_delay - (time.perf_counter_ns() - time_reference)) * 1e-9):
            time_reference = time.perf_counter_ns()
            grab_frame_method
            grabbed_frames_count += 1
            fps_counter += 1
            if ret:
                self.set_value(["Grabber", "Frame"], cv2.cvtColor(cv_img, cv2.COLOR_BGR2RGB))

            if (time.perf_counter_ns() - fps_counter_reference) >= 1e9:
                self.set_raw_value(ATTR_FRAMERATE, fps_counter)
                self.set_raw_value(ATTR_FRAMES_GRABBED, grabbed_frames_count)
                fps_counter = 0
                fps_counter_reference = time.perf_counter_ns()
        """
        raise NotImplementedError

    def set_max_broadcast_rate(self, value):
        self[ATTR_OUTPUT][ATTR_FRAME][ALLOW_BROADCAST] = value > 0
        if value > 0:
            self[ATTR_OUTPUT][ATTR_FRAME].min_broadcast_timeout = (1e9 / value) * 0.95

    def set_framerate(self, value):
        self.frame_delay = 1e9 / value

    def handle_configuration(self):
        start_at = time.time()
        self._config_attributes()
        self.set_status(STATUS_CONFIGURED)
        self.logger.info(
            "Device configuration took {} sec.".format(time.time() - start_at)
        )

    @expose_method()
    def stop(self):
        self.stopped.set()
        self.set_status(STATUS_IDLE)

    def close(self):
        self.stop()
        Device.close(self)



""" Example of yaml configuration
Smaract0: &Smaract0 !Device:kamzik3.devices.deviceSmaractMcs1.DeviceSmaractMcs2
    device_id: Smaract0
    host: 192.168.83.91
Smaract0_axis_0: &Smaract0_axis_0 !Device:kamzik3.devices.deviceSmaractMcs1.DeviceSmaractMcs1Channe2
    device: *Smaract0
    channel: 0
    device_id: MotorAxis0
    config:
      attributes:
        !!python/tuple [Sensor type, Value]: 1
        !!python/tuple [Position +limit, Value]: 15mm
        !!python/tuple [Position -limit, Value]: -15mm
        !!python/tuple [Position, Tolerance]: [10nm, 10nm]


Positioner codes:
SL...S1SS 300 Linear positioner, single piezo element
SL...D1SS 301 Linear positioner, double piezo element
SL...S1SC1 303 Linear positioner, single piezo element
SL...D1SC1 304 Linear positioner, double piezo element
SL...D1SC2 307 Linear positioner, double piezo element
SR...S1S5S 309 Small rotary positioner S single mark
SR...S1S6S 312 Rotary positioner S single mark
SR...D1S6S 313 Rotary positioner, double piezo element
SR...D1S7S 316 Rotary positioner, double piezo element
SR...T1S8S 320 Large rotary positioner S single mark
SG...D1S1S 325 Goniometer, 60.5mm radius, double
SG...D1S2S 328 Goniometer, 77.5mm radius, double
SL...S1LE 342 Linear positioner, single piezo element
SL...D1LE 343 Linear positioner, double piezo element
SL...S1LS 345 Linear positioner, single piezo element
SL...D1LS 346 Linear positioner, double piezo element
SL...S1LC1 348 Linear positioner, single piezo element
SL...D1LC1 349 Linear positioner, double piezo element
SR...S1L2S 354 Rotary positioner L single mark
SR...D1L2S 355 Rotary positioner, double piezo element
SL...S1ME 357 Linear positioner, single piezo element
SL...D1ME 358 Linear positioner, double piezo element
SL...S1P1E 360 Linear positioner, double piezo element
SL...D1P1E 361 Linear positioner, double piezo element
SG...S1M1E 363 Goniometer, 60.5mm radius, single
SG...S1M2E 366 Goniometer, 77.5mm radius, single
SG...D1L1S 381 Goniometer, 60.5mm radius, double
SG...D1L2S 383 Goniometer, 77.5mm radius, double
SG...D1L1E 387 Goniometer, 60.5mm radius, double
SG...D1L2E 389 Goniometer, 77.5mm radius, double
SH...A1SS 395 High load table, single piezo element S single mark
"""
import re
import time

import numpy as np
from bidict import frozenbidict
from numbers import Real
from pint.errors import UndefinedUnitError
from typing import List, Callable, Optional

import kamzik3
from kamzik3 import DeviceError, CommandFormatException, DeviceUnitError, units
from kamzik3.constants import *
from kamzik3.devices.deviceChannel import DeviceChannel
from kamzik3.devices.deviceSocket import DeviceSocket
from kamzik3.snippets.snippetsDecorators import expose_method
from kamzik3.snippets.snippetsUnits import device_units


class DeviceSmaractMcs2(DeviceSocket):
    response_timeout = 3000
    terminator = b"\r\n"
    error_pattern = re.compile(r'([-0-9]+),"([\w ]+)"')
    sensor_modes = frozenbidict({"Disabled": 0, "Enabled": 1, "Power safe": 2})
    move_modes = frozenbidict(
        {
            "Absolute": 0,
            "Relative": 1,
            "Scan absolute": 2,
            "Scan relative": 3,
            "Step": 4,
        }
    )
    safe_direction_modes = frozenbidict({"Forward": 0, "Backward": 1})
    emergency_stop_modes = frozenbidict(
        {"Normal": 0, "Restricted": 1, "Auto release": 2}
    )

    def __init__(self, host, port=55551, device_id=None, config=None):
        super().__init__(host, port, device_id, config)

    def _init_attributes(self):
        super()._init_attributes()
        self.create_attribute(ATTR_STATE, readonly=True, default_type=np.uint16)
        self.create_attribute(ATTR_SERIAL_NUMBER, readonly=True)
        self.create_attribute(ATTR_CHANNELS, readonly=True, default_type=np.uint16)
        self.create_attribute(ATTR_MODULES, readonly=True, default_type=np.uint16)
        self.create_attribute(ATTR_HAND_TOOL, readonly=True, default_type=bool)
        self.create_attribute(ATTR_MOVEMENT_LOCKED, readonly=True, default_type=bool)
        self.create_attribute(
            ATTR_INTERNAL_COMM_FAILURE, readonly=True, default_type=bool
        )
        self.create_attribute(ATTR_STREAMING, readonly=True, default_type=bool)
        self.create_attribute(
            ATTR_EMERGENCY_STOP,
            default_type=self.emergency_stop_modes.keys(),
            set_function=self.set_emergency_stop_mode,
        )

    def command(self, command, callback=None, with_token=False, returning=True):
        if not self.valid_command_format(command):
            raise CommandFormatException("Command '{}' form is invalid".format(command))

        returning = True
        command += ":SYST:ERR:NEXT?\r\n"
        return super().command(command, callback, with_token, returning)

    def poll_command(self, command, interval):
        if not self.valid_command_format(command):
            raise CommandFormatException("Command '{}' form is invalid".format(command))

        command += ":SYST:ERR:NEXT?\r\n"
        super().poll_command(command, interval)

    def remove_poll_command(self, command, interval):
        if not self.valid_command_format(command):
            raise CommandFormatException("Command '{}' form is invalid".format(command))

        command += ":SYST:ERR:NEXT?\r\n"
        super().remove_poll_command(command, interval)

    def handle_configuration(self):

        start_at = time.time()

        def _finish_configuration(*_, **__):
            self._config_commands()
            self._config_attributes()
            self.set_status(STATUS_CONFIGURED)
            self.logger.info(
                "Device configuration took {} sec.".format(time.time() - start_at)
            )

        # Check for possible errors
        self.get_errors_count()
        # Get number of channels
        self.command(":DEV:NOCH?\r\n")
        # Get serial number
        self.command(":DEV:SNUM?\r\n")
        # Get number of modules
        self.command(":DEV:NOBM?\r\n", callback=_finish_configuration)
        # Get emergency stop mode
        self.command(":DEV:EST:MODE?\r\n")
        self.start_polling()

    def handle_configuration_event(self):
        try:
            start_at = time.time()
            self.set_status(STATUS_CONFIGURING)
            self.handle_configuration()
            self.logger.info(
                "Device configuration took {} sec.".format(time.time() - start_at)
            )
        except DeviceError:
            self.logger.exception("Error during configuration")

    def valid_command_format(self, command):
        return command is not None and command[-2:] == "\r\n"

    def handle_readout(self, readout_buffer):
        is_error_message = self.error_pattern.match(readout_buffer[0])
        command, output, callback, token = super().handle_readout(readout_buffer)
        if is_error_message is not None:
            if command == ":SYST:ERR:NEXT?\r\n" and int(is_error_message.group(1)) != 0:
                self.logger.error(readout_buffer[0])
            elif int(is_error_message.group(1)) != 0:
                self.logger.error(readout_buffer[0])

            if callback is not None:
                self.handle_readout_callback(callback, command, readout_buffer)
            if token:
                kamzik3.session.publisher.push_message(
                    self.device_id, [command, readout_buffer], token
                )
                for observer in self._observers[:]:
                    if isinstance(observer, DeviceSmaractMcs2Channel):
                        kamzik3.session.publisher.push_message(
                            observer.device_id, [command, readout_buffer], token
                        )
            return
        else:
            command = command.replace(":SYST:ERR:NEXT?\r\n", "")
            self.commands_buffer.appendleft(
                ((":SYST:ERR:NEXT?\r\n", None, None, True), time.time())
            )

        command = command.strip()

        if command == ":DEV:NOCH?":
            self.set_offsetted_value(ATTR_CHANNELS, int(output))
        elif command == ":DEV:SNUM?":
            self.set_offsetted_value(ATTR_SERIAL_NUMBER, output.replace('"', ""))
        elif command == ":DEV:NOBM?":
            self.set_offsetted_value(ATTR_MODULES, int(output))
            for i in range(int(output)):
                module_name = "Module{}".format(i)
                self.create_attribute(
                    ATTR_STATE, module_name, readonly=True, default_type=np.uint16
                )
                self.create_attribute(
                    ATTR_TEMPERATURE,
                    module_name,
                    readonly=True,
                    default_type=np.float32,
                    decimals=2,
                    unit="degC",
                )
                self.create_attribute(
                    ATTR_SENSOR_PRESENT, module_name, readonly=True, default_type=bool
                )
                self.create_attribute(
                    ATTR_BOOSTER_PRESENT, module_name, readonly=True, default_type=bool
                )
                self.create_attribute(
                    ATTR_ADJUSTMENT_ACTIVE,
                    module_name,
                    readonly=True,
                    default_type=bool,
                )
                self.create_attribute(
                    ATTR_IOM_PRESENT, module_name, readonly=True, default_type=bool
                )
                self.create_attribute(
                    ATTR_INTERNAL_COMM_FAILURE,
                    module_name,
                    readonly=True,
                    default_type=bool,
                )
                self.create_attribute(
                    ATTR_HIGH_VOLTAGE_FAILURE,
                    module_name,
                    readonly=True,
                    default_type=bool,
                )
                self.create_attribute(
                    ATTR_HIGH_VOLTAGE_OVERLOAD,
                    module_name,
                    readonly=True,
                    default_type=bool,
                )
                self.create_attribute(
                    ATTR_OVER_TEMPERATURE, module_name, readonly=True, default_type=bool
                )

                self.poll_command(":MOD{}:STAT?\r\n".format(i), 2000)
                self.poll_command(":MOD{}:TEMP?\r\n".format(i), 2000)
        elif command == ":DEV:STAT?":
            output = int(output)
            self.set_offsetted_value(ATTR_STATE, output)
            self.set_offsetted_value(ATTR_HAND_TOOL, bool(output % 0x0001))
            self.set_offsetted_value(ATTR_MOVEMENT_LOCKED, bool(output % 0x0002))
            self.set_offsetted_value(ATTR_INTERNAL_COMM_FAILURE, bool(output % 0x0100))
            self.set_offsetted_value(ATTR_STREAMING, bool(output % 0x1000))
        elif command == ":DEV:EST:MODE?":
            output = int(output)
            self.set_offsetted_value(
                ATTR_EMERGENCY_STOP, self.emergency_stop_modes.inverse[output]
            )
        elif command == ":SYST:ERR:COUN?":
            if int(output) > 0:
                self.get_next_error()
        elif command == ":SYST:ERR:NEXT?":
            self.logger.error(output)
            self.get_errors_count()
        elif command[:4] == ":MOD":
            _, module, attribute = command.split(":")
            module_name = "Module{}".format(module[3:])
            if attribute == "TEMP?":
                self.set_offsetted_value((module_name, ATTR_TEMPERATURE), float(output))
            elif attribute == "STAT?":
                output = int(output)
                self.set_offsetted_value((module_name, ATTR_STATE), output)
                self.set_offsetted_value(
                    (module_name, ATTR_SENSOR_PRESENT), bool(output & 0x0001)
                )
                self.set_offsetted_value(
                    (module_name, ATTR_BOOSTER_PRESENT), bool(output & 0x0002)
                )
                self.set_offsetted_value(
                    (module_name, ATTR_ADJUSTMENT_ACTIVE), bool(output & 0x0004)
                )
                self.set_offsetted_value(
                    (module_name, ATTR_IOM_PRESENT), bool(output & 0x0008)
                )
                self.set_offsetted_value(
                    (module_name, ATTR_INTERNAL_COMM_FAILURE), bool(output & 0x0100)
                )
                self.set_offsetted_value(
                    (module_name, ATTR_HIGH_VOLTAGE_FAILURE), bool(output & 0x1000)
                )
                self.set_offsetted_value(
                    (module_name, ATTR_HIGH_VOLTAGE_OVERLOAD), bool(output & 0x2000)
                )
                self.set_offsetted_value(
                    (module_name, ATTR_OVER_TEMPERATURE), bool(output & 0x4000)
                )
        elif command[:5] == ":CHAN":
            chunks = command.split(":")
            channel, attribute = chunks[1], chunks[-1]
            channel = int(channel[4:])
            if attribute == "STAT?":
                attribute, output = ATTR_STATE, int(output)
            elif attribute == "CURR?":
                attribute, output = ATTR_POSITION, int(output)
            elif attribute == "TARG?":
                attribute, output = ATTR_TARGET_POSITION, int(output)
            elif attribute == "TEMP?":
                attribute, output = ATTR_TEMPERATURE, float(output)
            elif attribute == "UNIT?":
                attribute, output = ATTR_ENCODER_BASE_UNIT, int(output)
            elif attribute == "RES?":
                attribute, output = ATTR_ENCODER_RESOLUTION, int(output)
            elif attribute == "NAME?":
                attribute, output = ATTR_POSITIONER_TYPE, output.replace('"', "")
            elif attribute == "CODE?":
                attribute, output = ATTR_POSITIONER_CODE, int(output)
            elif attribute == "MIN?":
                attribute, output = ATTR_POSITION_LOWER_LIMIT, int(output)
            elif attribute == "MAX?":
                attribute, output = ATTR_POSITION_UPPER_LIMIT, int(output)
            elif attribute == "FREQ?":
                attribute, output = ATTR_STEP_FREQUENCY, int(output)
            elif attribute == "AMPL?":
                attribute, output = ATTR_STEP_AMPLITUDE, int(output)
            elif attribute == "VEL?":
                attribute, output = ATTR_VELOCITY, int(output)
            elif attribute == "HOLD?":
                attribute, output = ATTR_HOLDING_TIME, int(output)
            elif attribute == "ACC?":
                attribute, output = ATTR_ACCELERATION, int(output)
            elif attribute == "MODE?":
                attribute, output = (
                    ATTR_SENSOR_MODE,
                    self.sensor_modes.inverse[int(output)],
                )
            elif attribute == "MMOD?":
                attribute, output = ATTR_MOVE_MODE, self.move_modes.inverse[int(output)]
            elif attribute == "SDIR?":
                attribute, output = (
                    ATTR_SAFE_DIRECTION,
                    self.safe_direction_modes.inverse[int(output)],
                )
            elif attribute == "OPT?":
                attribute, output = (
                    ATTR_REFERENCE_DIRECTION,
                    self.safe_direction_modes.inverse[int(output)],
                )
            else:
                return
            self.notify((channel, attribute), output)

        if callback is not None:
            self.handle_readout_callback(callback, command, readout_buffer)
        if token:
            kamzik3.session.publisher.push_message(
                self.device_id, [command, readout_buffer], token
            )
            for observer in self._observers[:]:
                if isinstance(observer, DeviceSmaractMcs2Channel):
                    kamzik3.session.publisher.push_message(
                        observer.device_id, [command, readout_buffer], token
                    )

    def start_polling(self):
        super().start_polling()
        self.poll_command(":DEV:STAT?\r\n", 1000)

    def get_errors_count(self):
        self.command(":SYST:ERR:COUN?\r\n")

    def get_next_error(self):
        self.command(":SYST:ERR:NEXT?\r\n")

    def set_emergency_stop_mode(self, value, callback=None):
        self.logger.info("Set emergency stop mode to {}".format(value))
        value = self.emergency_stop_modes[value]
        # Set this mode as a default
        self.command(
            ":DEV:EST:MODE:DEF {}\r\n".format(value), callback, with_token=True
        )
        return self.command(
            ":DEV:EST:MODE {}\r\n".format(value), callback, with_token=True
        )


class DeviceSmaractMcs2Channel(DeviceChannel):
    sensor_modes = DeviceSmaractMcs2.sensor_modes
    move_modes = DeviceSmaractMcs2.move_modes
    safe_direction_modes = DeviceSmaractMcs2.safe_direction_modes

    def _init_attributes(self):
        super()._init_attributes()
        self.create_attribute(ATTR_STATE, readonly=True, default_type=np.uint16)
        self.create_attribute(
            ATTR_TEMPERATURE,
            readonly=True,
            default_type=np.float16,
            decimals=2,
            unit="degC",
        )
        self.create_attribute(ATTR_MOVING, readonly=True, default_type=bool)
        self.create_attribute(
            ATTR_POSITION,
            readonly=False,
            default_type=np.int64,
            set_function=self._move_absolute,
            set_value_when_set_function=False,
        )
        self.create_attribute(
            ATTR_TARGET_POSITION,
            default_type=np.int64,
            set_function=self._move_absolute,
            set_value_when_set_function=False,
        )
        self.create_attribute(ATTR_CLOSED_LOOP_ACTIVE, readonly=True, default_type=bool)
        self.create_attribute(ATTR_CALIBRATING, readonly=True, default_type=bool)
        self.create_attribute(ATTR_REFERENCING, readonly=True, default_type=bool)
        self.create_attribute(ATTR_HOLDING, readonly=True, default_type=bool)
        self.create_attribute(ATTR_MOVE_DELAYED, readonly=True, default_type=bool)
        self.create_attribute(ATTR_SENSOR_PRESENT, readonly=True, default_type=bool)
        self.create_attribute(ATTR_CALIBRATED, readonly=True, default_type=bool)
        self.create_attribute(ATTR_REFERENCED, readonly=True, default_type=bool)
        self.create_attribute(ATTR_END_STOP_REACHED, readonly=True, default_type=bool)
        self.create_attribute(
            ATTR_RANGE_LIMIT_REACHED, readonly=True, default_type=bool
        )
        self.create_attribute(
            ATTR_FOLLOWING_LIMIT_REACHED, readonly=True, default_type=bool
        )
        self.create_attribute(ATTR_MOVEMENT_FAILED, readonly=True, default_type=bool)
        self.create_attribute(ATTR_STREAMING, readonly=True, default_type=bool)
        self.create_attribute(ATTR_OVER_TEMPERATURE, readonly=True, default_type=bool)
        self.create_attribute(ATTR_INDEX_MARK, readonly=True, default_type=bool)
        self.create_attribute(ATTR_ENCODER_BASE_UNIT, readonly=True)
        self.create_attribute(
            ATTR_ENCODER_RESOLUTION, readonly=True, default_type=np.int8
        )
        self.create_attribute(
            ATTR_POSITIONER_CODE,
            readonly=False,
            default_type=np.int8,
            min_value=0,
            max_value=500,
            set_function=self.set_positioner_code,
        )
        self.create_attribute(ATTR_POSITIONER_TYPE, readonly=True)
        self.create_attribute(
            ATTR_POSITION_LOWER_LIMIT,
            default_type=np.int64,
            set_function=self.set_lower_position_limit,
        )
        self.create_attribute(
            ATTR_POSITION_UPPER_LIMIT,
            default_type=np.int64,
            set_function=self.set_upper_position_limit,
        )
        self.create_attribute(
            ATTR_CONFIGURED_LIMITS,
            default_type=TYPE_LIST,
            default_value=[None, None],
            set_function=self._check_limits,
        )
        self.create_attribute(
            ATTR_STEP_FREQUENCY,
            default_type=np.uint32,
            min_value=1,
            max_value=20e3,
            unit="Hz",
            set_function=self.set_step_frequency,
        )
        self.create_attribute(
            ATTR_STEP_AMPLITUDE,
            default_type=np.uint32,
            min_value=1,
            max_value=65535,
            set_function=self.set_step_amplitude,
        )
        self.create_attribute(
            ATTR_ACCELERATION,
            default_type=np.int64,
            min_value=0,
            max_value=100e9,
            set_function=self.set_acceleration,
        )
        self.create_attribute(
            ATTR_VELOCITY,
            default_type=np.int64,
            min_value=0,
            max_value=10e12,
            set_function=self.set_velocity,
        )
        self.create_attribute(
            ATTR_SENSOR_MODE,
            default_type=self.sensor_modes.keys(),
            set_function=self.set_sensor_mode,
        )
        self.create_attribute(
            ATTR_MOVE_MODE,
            default_type=self.move_modes.keys(),
            set_function=self.set_move_mode,
        )
        self.create_attribute(
            ATTR_SAFE_DIRECTION,
            default_type=self.safe_direction_modes.keys(),
            set_function=self.set_safe_direction,
        )
        self.create_attribute(
            ATTR_REFERENCE_DIRECTION,
            default_type=self.safe_direction_modes.keys(),
            set_function=self.set_reference_direction,
        )
        self.create_attribute(
            ATTR_HOLDING_TIME,
            default_type=np.int32,
            min_value=-1,
            unit="ms",
            set_function=self.set_holding_time,
        )
        self.create_attribute(ATTR_CLOSE_LOOP_UNIT, display=False)
        self.create_attribute(
            ATTR_OPEN_LOOP_FACTOR,
            default_value=[1, 1],
            default_type=TYPE_LIST,
            display=False,
        )
        self.create_attribute(
            ATTR_MOVE_METHOD,
            default_value=CLOSE_LOOP,
            default_type=[OPEN_LOOP, CLOSE_LOOP],
            set_function=self.set_move_method,
        )

    def _check_limits(self, limits: List[Real]) -> None:
        """
        Check if the limits are correctly configured.

        :param limits: list of two real numbers [lower_limit, upper_limit], None for a
         is limit is allowed.
        """
        if limits[0] == limits[1] or (  # limits could be a number or None
            all(val is not None for val in limits) and limits[1] < limits[0]
        ):
            self.logger.info("Invalid configuration for limits, defaulting to -/+inf")
            self.set_attribute((ATTR_POSITION, MIN), -np.inf)
            self.set_attribute((ATTR_POSITION, MAX), np.inf)
            self.set_value(ATTR_POSITION_LOWER_LIMIT, -np.inf)
            self.set_value(ATTR_POSITION_UPPER_LIMIT, np.inf)

    def _update_configured_limits(self, _):
        """
        Callback which updates ATTR_CONFIGURED_LIMITS.

        Another callback is attached to ATTR_CONFIGURED_LIMITS, which allows to check
        if limits are compatible with each other.
        """
        lower_limit = self.get_value(ATTR_POSITION_LOWER_LIMIT, VALUE)
        upper_limit = self.get_value(ATTR_POSITION_UPPER_LIMIT, VALUE)
        self.set_value(ATTR_CONFIGURED_LIMITS, [lower_limit, upper_limit])

    def _check_move_mode(self, value):
        """
        Check the compatibility of the configured move mode with mode method.

        It is important to check the value of ATTR_MOVE_MODE here, because if the user
        uses the controller to move the stage, the device switches to 'Relative' move
        mode without Kamzik knowing it.
        """
        if self.get_value(ATTR_MOVE_METHOD) == OPEN_LOOP and value != "Step":
            self.set_value(ATTR_MOVE_MODE, "Step")
        if self.get_value(ATTR_MOVE_METHOD) == CLOSE_LOOP and value != "Absolute":
            self.set_value(ATTR_MOVE_MODE, "Absolute")

    def config_attributes(self):
        """
        Check the attributes configuration.

        The base resolution is by default not writable with the MCS2 controller.
        Therefore the measured position will always be in the default unit (pm or ndeg).
        One must prevent the user to define a wrong resolution in the config, since this
        would not apply to the device.
        """
        if ("Position", "Unit") in self.config["attributes"]:
            self.logger.info(
                "Ignoring '[ Position, Unit ]: "
                f"{self.config['attributes'][('Position', 'Unit')]}' from config file"
            )
            del self.config["attributes"][("Position", "Unit")]

        self._config_attributes()

    def handle_configuration(self):
        if self.configured:
            return
        self.attach_attribute_callback(
            ATTR_POSITION_LOWER_LIMIT,
            callback=self._update_configured_limits,
            key_filter=VALUE,
        )
        self.attach_attribute_callback(
            ATTR_POSITION_UPPER_LIMIT,
            callback=self._update_configured_limits,
            key_filter=VALUE,
        )
        self.attach_attribute_callback(
            ATTR_MOVE_MODE,
            callback=self._check_move_mode,
            key_filter=VALUE,
        )

        start_at = time.time()

        def _finish_configuration(*_, **__):
            self._config_commands()
            self.config_attributes()
            self.configured = True
            self.start_polling()
            self.set_status(STATUS_CONFIGURED)
            self.logger.info(
                "Device configuration took {} sec.".format(time.time() - start_at)
            )

        self.connected = True
        self.get_sensor_mode()
        self.get_encoder_unit()
        self.get_position_limit_range()
        self.get_positioner_configuration()
        self.get_step_frequency()
        self.get_step_amplitude()
        self.get_acceleration()
        self.get_velocity()
        self.get_move_mode()
        self.get_holding_time()
        self.get_reference_direction()
        self.command(
            ":CHAN{}:SDIR?\r\n".format(self.channel), callback=_finish_configuration
        )

    def start_polling(self):
        super().start_polling()
        self.poll_command(":CHAN{}:STAT?\r\n".format(self.channel), 250)
        self.poll_command(":CHAN{}:MMOD?\r\n".format(self.channel), 1000)
        self.poll_command(":CHAN{}:POS:CURR?\r\n".format(self.channel), 250)
        self.poll_command(":CHAN{}:POS:TARG?\r\n".format(self.channel), 250)
        self.poll_command(":CHAN{}:TEMP?\r\n".format(self.channel), 10000)

    def stop_polling(self):
        self.remove_poll_command(":CHAN{}:STAT?\r\n".format(self.channel), 250)
        self.remove_poll_command(":CHAN{}:MMOD?\r\n".format(self.channel), 1000)
        self.remove_poll_command(":CHAN{}:POS:CURR?\r\n".format(self.channel), 250)
        self.remove_poll_command(":CHAN{}:POS:TARG?\r\n".format(self.channel), 250)
        self.remove_poll_command(":CHAN{}:TEMP?\r\n".format(self.channel), 10000)

    def subject_update(self, key, value, subject):
        super().subject_update(key, value, subject)

        if self.connected and isinstance(key, tuple) and key[0] == self.channel:
            attribute = key[1]

            if attribute == ATTR_STATE and self.get_value(ATTR_STATE) != value:
                if self.get_value(ATTR_SENSOR_PRESENT) != bool(value & 0x0020):
                    if bool(value & 0x0020):
                        self.poll_command(
                            ":CHAN{}:POS:CURR?\r\n".format(self.channel), 250
                        )
                        self.poll_command(
                            ":CHAN{}:POS:TARG?\r\n".format(self.channel), 250
                        )
                    else:
                        self.remove_poll_command(
                            ":CHAN{}:POS:CURR?\r\n".format(self.channel), 250
                        )
                        self.remove_poll_command(
                            ":CHAN{}:POS:TARG?\r\n".format(self.channel), 250
                        )

                self.set_offsetted_value(ATTR_MOVING, bool(value & 0x0001))
                self.set_offsetted_value(ATTR_CLOSED_LOOP_ACTIVE, bool(value & 0x0002))
                self.set_offsetted_value(ATTR_CALIBRATING, bool(value & 0x0004))
                self.set_offsetted_value(ATTR_REFERENCING, bool(value & 0x0008))
                self.set_offsetted_value(ATTR_MOVE_DELAYED, bool(value & 0x0100))
                self.set_offsetted_value(ATTR_SENSOR_PRESENT, bool(value & 0x0020))
                self.set_offsetted_value(ATTR_CALIBRATED, bool(value & 0x0040))
                self.set_offsetted_value(ATTR_REFERENCED, bool(value & 0x0080))
                self.set_offsetted_value(ATTR_END_STOP_REACHED, bool(value & 0x0100))
                self.set_offsetted_value(ATTR_RANGE_LIMIT_REACHED, bool(value & 0x0200))
                self.set_offsetted_value(
                    ATTR_FOLLOWING_LIMIT_REACHED, bool(value & 0x0400)
                )
                self.set_offsetted_value(ATTR_MOVEMENT_FAILED, bool(value & 0x0800))
                self.set_offsetted_value(ATTR_STREAMING, bool(value & 0x1000))
                self.set_offsetted_value(ATTR_OVER_TEMPERATURE, bool(value & 0x4000))
                self.set_offsetted_value(ATTR_INDEX_MARK, bool(value & 0x8000))

                holding = (
                    bool(value & 0x0004)
                    and bool(value & 0x0008)
                    and bool(value & 0x0008)
                    and bool(value & 0x0002)
                    and not bool(value & 0x0001)
                )
                self.set_offsetted_value(ATTR_HOLDING, holding)

                if bool(value & 0x0001):
                    self.set_status(STATUS_BUSY)
                else:
                    self.set_status(STATUS_IDLE)
            elif attribute == ATTR_SENSOR_MODE:
                if value == "Disabled":
                    self.remove_poll_command(
                        ":CHAN{}:POS:CURR?\r\n".format(self.channel), 250
                    )
                    self.remove_poll_command(
                        ":CHAN{}:POS:TARG?\r\n".format(self.channel), 250
                    )
                else:
                    self.poll_command(":CHAN{}:POS:CURR?\r\n".format(self.channel), 250)
                    self.poll_command(":CHAN{}:POS:TARG?\r\n".format(self.channel), 250)
            elif attribute == ATTR_ENCODER_BASE_UNIT:
                if value == 2:
                    value = "m"
                else:
                    value = "deg"
            elif attribute == ATTR_ENCODER_RESOLUTION:
                unit = self.get_value(ATTR_ENCODER_BASE_UNIT)
                prefix = {-12: "p", -9: "n", -6: "u", -3: "m"}
                base_position_unit = prefix.get(value, "") + unit

                self.set_attribute((ATTR_TARGET_POSITION, UNIT), base_position_unit)
                self.set_attribute(
                    (ATTR_POSITION_UPPER_LIMIT, UNIT), base_position_unit
                )
                self.set_attribute(
                    (ATTR_POSITION_LOWER_LIMIT, UNIT), base_position_unit
                )
                self.set_attribute(
                    (ATTR_ACCELERATION, UNIT), "{}/s".format(base_position_unit)
                )
                self.set_attribute(
                    (ATTR_VELOCITY, UNIT), "{}/s".format(base_position_unit)
                )
                self.set_attribute((ATTR_POSITION, UNIT), base_position_unit)
            elif attribute == ATTR_POSITION_LOWER_LIMIT:
                self.set_attribute((ATTR_POSITION, MIN), value)
            elif attribute == ATTR_POSITION_UPPER_LIMIT:
                self.set_attribute((ATTR_POSITION, MAX), value)
            self.set_value(ATTR_LATENCY, self.device.get_value(ATTR_LATENCY))
            self.set_offsetted_value(attribute, value)

    def enable_close_loop(self):
        if self.position_attribute_copy is not None:
            self.set_attribute(
                (ATTR_POSITION, UNIT), self.position_attribute_copy[UNIT]
            )
            self.set_attribute(
                (ATTR_POSITION, OFFSET), self.position_attribute_copy[OFFSET]
            )
            self.set_attribute(
                (ATTR_POSITION, FACTOR), self.position_attribute_copy[FACTOR]
            )
            self.set_value(ATTR_MOVE_MODE, "Absolute")
            self[ATTR_POSITION].set_function = self._move_absolute
            self.position_attribute_copy = None

            # self.poll_command(":CHAN{}:POS:CURR?\r\n".format(self.channel), 250)
            # self.poll_command(":CHAN{}:POS:TARG?\r\n".format(self.channel), 250)

    def enable_open_loop(self):
        self.position_attribute_copy = self[ATTR_POSITION].attribute_copy()
        default_unit = self.get_value(ATTR_CLOSE_LOOP_UNIT)

        self.set_attribute(
            (ATTR_POSITION, UNIT),
            default_unit
            if default_unit is not None
            else self.position_attribute_copy[UNIT],
        )
        self.set_attribute((ATTR_POSITION, OFFSET), 0)
        self.set_attribute((ATTR_POSITION, FACTOR), 1)

        if self.get_value(ATTR_POSITION) is None:
            self.set_offsetted_value(ATTR_POSITION, 0)
        self[ATTR_POSITION].set_function = self._move_steps
        self.set_value(ATTR_MOVE_MODE, "Step")

        # self.remove_poll_command(":CHAN{}:POS:CURR?\r\n".format(self.channel), 250)
        # self.remove_poll_command(":CHAN{}:POS:TARG?\r\n".format(self.channel), 250)

    def set_move_method(self, value):
        if value == OPEN_LOOP:
            self.enable_open_loop()
        elif value == CLOSE_LOOP:
            self.enable_close_loop()

    @expose_method()
    def stop(self, callback=None):
        self.logger.info("Stop movement")
        return self.command(
            ":STOP{}\r\n".format(self.channel), callback, with_token=True
        )

    @expose_method()
    def calibrate(self, callback=None):
        self.logger.info("Calibrate")
        return self.command(
            ":CAL{}\r\n".format(self.channel), callback, with_token=True
        )

    @expose_method()
    def find_reference(self, callback=None):
        self.logger.info("Reference")
        return self.command(
            ":REF{}\r\n".format(self.channel), callback, with_token=True
        )

    @expose_method({"step": ATTR_POSITION})
    def move_relative(self, step):
        """Move the stage relatively to the current position."""
        try:
            step = device_units(self, ATTR_POSITION, step).m
            self.logger.info("Move by relative position {}".format(step))
            if self.get_value(ATTR_MOVE_MODE) == "Absolute":
                # add the step to the non-offsetted current position
                step = self.get_raw_value(ATTR_POSITION) + step
            return self.set_value(ATTR_TARGET_POSITION, step)
        except (UndefinedUnitError, ValueError) as e:
            self.logger.exception(e)
            raise DeviceUnitError(e)

    @expose_method({"position": ATTR_POSITION})
    def move_absolute(self, position):
        """
        Move the stage to the defined absolute position.

        The position from the GUI includes the offset, which needs to be subtracted
        before calling _move_absolute (which sends the command to the device).
        """
        try:
            position = device_units(self, ATTR_POSITION, position).m
            unit = self.get_value(ATTR_POSITION, key=UNIT)
            self.logger.info(
                "Move to position including offset {} {}".format(position, unit)
            )
            position = position - self.get_value(ATTR_POSITION, key=OFFSET)
            return self.set_value(ATTR_TARGET_POSITION, position)
        except (UndefinedUnitError, ValueError) as e:
            self.logger.exception(e)
            raise DeviceUnitError(e)

    def _move_absolute(self, value, callback=None):
        """
        Move in absolute position of the device.

        :param value: physical position = GUI_position/factor - offset
        """
        if self.get_value(ATTR_POSITION) is not None:
            self.logger.info("Move to absolute position {}".format(value))
            token = self.command(
                ":MOVE{} {}\r\n".format(self.channel, value), callback, with_token=True
            )
            return token
        return None

    def _move_steps(self, steps: float, callback: Optional[Callable] = None) -> int:
        """
        Callback method for open loop control.

        The amplitude of the movement is unitless, in steps. The n_factor and p_factor
        are user-friendly conversion factors corresponding to the estimated number of
        steps per degree or per millimeter (in the positive and negative directions).

        :param steps: number of steps to move, expressed in the unit of ATTR_POSITION
        :param callback: optional callback
        :return: the token
        """
        if self.get_value(ATTR_MOVE_MODE) != "Step":
            self.set_value(ATTR_MOVE_MODE, "Step")

        # convert steps to the user-friendly unit deg or mm
        unit = self.get_value(ATTR_POSITION, UNIT)
        if unit in {"mdeg", "udeg", "ndeg", "pdeg"}:
            steps = units.Quantity(steps, unit).to("deg").m
        if unit in {"m", "um", "nm", "pm"}:
            steps = units.Quantity(steps, unit).to("mm").m

        n_factor, p_factor = self.get_value(ATTR_OPEN_LOOP_FACTOR)
        factor = n_factor if steps < 0 else p_factor
        steps = int(np.rint(steps * factor))
        if steps == 0:
            self.logger.info(
                f"Move by {steps} steps: minimum angular step = {1/factor:.3f} deg"
            )
        else:
            self.logger.info(f"Move by {steps} steps")
        token = self.command(
            ":MOVE{} {}\r\n".format(self.channel, steps), callback, with_token=True
        )
        return int(token)

    @expose_method({"position": "New position"})
    def set_position(self, position=0):
        self.logger.info(f"Reset position to {position}")
        position = int(position)
        self.command(f":CHAN{self.channel}:POS {position}\r\n")

    def get_encoder_unit(self):
        self.command(":CHAN{}:TUN:BASE:UNIT?\r\n".format(self.channel))
        self.command(":CHAN{}:TUN:BASE:RES?\r\n".format(self.channel))

    def get_position_limit_range(self):
        self.command(":CHAN{}:RLIM:MAX?\r\n".format(self.channel))
        self.command(":CHAN{}:RLIM:MIN?\r\n".format(self.channel))

    def get_positioner_configuration(self):
        self.command(":CHAN{}:PTYP:NAME?\r\n".format(self.channel))
        self.command(":CHAN{}:PTYP:CODE?\r\n".format(self.channel))

    def get_step_frequency(self):
        self.command(":CHAN{}:STEP:FREQ?\r\n".format(self.channel))

    def get_step_amplitude(self):
        self.command(":CHAN{}:STEP:AMPL?\r\n".format(self.channel))

    def get_velocity(self):
        self.command(":CHAN{}:VEL?\r\n".format(self.channel))

    def get_acceleration(self):
        self.command(":CHAN{}:ACC?\r\n".format(self.channel))

    def get_sensor_mode(self):
        self.command(":CHAN{}:SENS:MODE?\r\n".format(self.channel))

    def get_move_mode(self):
        return self.command(":CHAN{}:MMOD?\r\n".format(self.channel))

    def get_safe_direction(self):
        self.command(":CHAN{}:SDIR?\r\n".format(self.channel))

    def get_reference_direction(self):
        self.command(":CHAN{}:REF:OPT?\r\n".format(self.channel))

    def get_holding_time(self):
        self.command(":CHAN{}:HOLD?\r\n".format(self.channel))

    def set_holding_time(self, value, callback=None):
        self.logger.info("Set holding time to {}".format(value))
        return self.command(
            ":CHAN{}:HOLD {}\r\n".format(self.channel, int(value)),
            callback,
            with_token=True,
        )

    # callback is unused, but I'm not sure if I can just disable it
    # or if it's still used dynamically
    # pylint: disable=unused-argument
    def set_upper_position_limit(self, value, callback=None):
        unit = self.get_value(ATTR_POSITION_UPPER_LIMIT, key=UNIT)
        try:
            set_value = device_units(self, ATTR_POSITION, units.Quantity(value, unit)).m
        except UndefinedUnitError:
            set_value = value
        self.logger.info("Set upper limit to {} {}".format(set_value, unit))
        self.set_attribute([ATTR_POSITION, MAX], set_value)

    # callback is unused, but I'm not sure if I can just disable it
    # or if it's still used dynamically
    # pylint: disable=unused-argument
    def set_lower_position_limit(self, value, callback=None):
        unit = self.get_value(ATTR_POSITION_LOWER_LIMIT, key=UNIT)
        try:
            set_value = device_units(self, ATTR_POSITION, units.Quantity(value, unit)).m
        except UndefinedUnitError:
            set_value = value
        self.logger.info("Set lower limit to {} {}".format(set_value, unit))
        self.set_attribute([ATTR_POSITION, MIN], set_value)

    def set_positioner_code(self, value, callback=None):
        self.logger.info("Set positioner  code to {}".format(value))
        token = self.command(
            ":CHAN{}:PTYP:CODE {}\r\n".format(self.channel, int(value)),
            callback,
            with_token=True,
        )
        self.get_positioner_configuration()
        self.get_encoder_unit()
        return token

    def set_step_frequency(self, value, callback=None):
        self.logger.info("Set step frequency to {} Hz".format(value))
        self.command(
            ":CHAN{}:MCLF:CURR {}\r\n".format(self.channel, int(value)),
            callback,
            with_token=True,
        )
        return self.command(
            ":CHAN{}:STEP:FREQ {}\r\n".format(self.channel, int(value)),
            callback,
            with_token=True,
        )

    def set_step_amplitude(self, value, callback=None):
        self.logger.info("Set step amplitude to {}".format(value))
        return self.command(
            ":CHAN{}:STEP:AMPL {}\r\n".format(self.channel, int(value)),
            callback,
            with_token=True,
        )

    def set_velocity(self, value, callback=None):
        self.logger.info("Set velocity to {}".format(value))
        return self.command(
            ":CHAN{}:VEL {}\r\n".format(self.channel, int(value)),
            callback,
            with_token=True,
        )

    def set_acceleration(self, value, callback=None):
        self.logger.info("Set acceleration to {}".format(value))
        return self.command(
            ":CHAN{}:ACC {}\r\n".format(self.channel, int(value)),
            callback,
            with_token=True,
        )

    def set_sensor_mode(self, value, callback=None):
        self.logger.info("Set sensor mode to {}".format(value))
        value = self.sensor_modes[value]
        out = self.command(
            ":CHAN{}:SENS:MODE {}\r\n".format(self.channel, value),
            callback,
            with_token=True,
        )
        self.get_sensor_mode()
        return out

    def set_move_mode(self, value, callback=None):
        self.logger.info("Set move mode to {}".format(value))
        if value == "Step":
            self.set_value(ATTR_MOVE_METHOD, OPEN_LOOP)
        else:
            self.set_value(ATTR_MOVE_METHOD, CLOSE_LOOP)

        value = self.move_modes[value]
        return self.command(
            ":CHAN{}:MMOD {}\r\n".format(self.channel, value), callback, with_token=True
        )

    def set_safe_direction(self, value, callback=None):
        self.logger.info("Set safe direction to {}".format(value))
        value = self.safe_direction_modes[value]
        return self.command(
            ":CHAN{}:SDIR {}\r\n".format(self.channel, value), callback, with_token=True
        )

    def set_reference_direction(self, value, callback=None):
        self.logger.info("Set reference direction to {}".format(value))
        value = self.safe_direction_modes[value]
        return self.command(
            ":CHAN{}:REF:OPT {}\r\n".format(self.channel, value),
            callback,
            with_token=True,
        )

    # callback is unused, but I'm not sure if I can just disable it
    # or if it's still used dynamically
    # pylint: disable=unused-argument
    @expose_method()
    def enable_holding(self, callback=None):
        """
        Here comes a short description of enable_holding.

        :param callback:
        """
        # enable holding
        self.set_value(ATTR_HOLDING_TIME, -1)
        # move by 0 to end holding
        self.set_value(ATTR_MOVE_MODE, "Relative")
        self.move_relative(device_units(self, ATTR_POSITION, 0))
        self.set_value(ATTR_MOVE_MODE, "Absolute")

    # callback is unused, but I'm not sure if I can just disable it
    # or if it's still used dynamically
    # pylint: disable=unused-argument
    @expose_method()
    def disable_holding(self, callback=None):
        """
        Here comes a short description of disable_holding.

        :param callback:
        """
        # disable holding
        self.set_value(ATTR_HOLDING_TIME, 0)
        # move by 0 to end holding
        self.set_value(ATTR_MOVE_MODE, "Relative")
        self.move_relative(device_units(self, ATTR_POSITION, 0))
        self.set_value(ATTR_MOVE_MODE, "Absolute")



import time

import numpy as np

from kamzik3.constants import *
from kamzik3.devices.attribute import Attribute
from kamzik3.devices.device import Device
from kamzik3.snippets.snippetsDecorators import expose_method
from kamzik3.snippets.snippetsTimer import CallbackTimer
from kamzik3.snippets.snippetsUnits import device_units


class DeviceShutter(Device):
    """
    General Shutter Device.
    To make use of it use this class as a parent class for any actual shutter.
    """

    def __init__(self, device_id=None, config=None):
        super().__init__(device_id, config)
        self.open_delay_timer = None
        self.close_delay_timer = None
        self.connect()

    def _init_attributes(self):
        """
        Attributes;
            ATTR_OPENED: Boolean flag showing open state of the Shutter
            ATTR_AUTO_MODE: Boolean switch, which enables Auto mode.
                            Auto mode is used by Macroserver to automatically operate the shutter.
            ATTR_OPEN_DELAY: Open delay in milliseconds
            ATTR_CLOSE_DELAY: Close delay in milliseconds
            ATTR_OPENED_AT: Showing to what percent shutter is Open
        """
        super()._init_attributes()
        self.create_attribute(
            ATTR_OPENED, default_value=False, readonly=True, default_type=np.bool
        )
        self.create_attribute(ATTR_AUTO_MODE, default_value=False, default_type=np.bool)
        self.create_attribute(
            ATTR_OPEN_DELAY,
            default_value=0,
            min_value=0,
            unit="ms",
            default_type=np.int32,
        )
        self.create_attribute(
            ATTR_CLOSE_DELAY,
            default_value=0,
            min_value=0,
            unit="ms",
            default_type=np.int32,
        )
        self.create_attribute(
            ATTR_OPENED_AT,
            default_value=0,
            readonly=True,
            default_type=np.float16,
            min_value=0,
            max_value=100,
            decimals=2,
            unit="%",
        )

    def handle_configuration(self):
        start_at = time.time()

        def _finish_configuration(*_, **__):
            self._config_commands()
            self._config_attributes()
            self.set_status(STATUS_CONFIGURED)
            self.logger.info(
                "Device configuration took {} sec.".format(time.time() - start_at)
            )

        _finish_configuration()

    def _open_shutter(self):
        """
        Privater method, which must be overloaded.
        Put code, which close the shutter based on Device specification here.
        """
        raise NotImplementedError()

    @expose_method()
    def open_shutter(self, clear_auto_mode=True):
        """
        This is general method which open the shutter.
        It's calling _open_shutter method either directly or as a callback when OPEN_DELAY is set.
        :param bool clear_auto_mode: Clear auto mode flag
        """
        self.clear_open_delay()
        self.set_value(ATTR_AUTO_MODE, not clear_auto_mode)

        open_delay = self.get_value(ATTR_OPEN_DELAY)
        if open_delay:
            self.set_status(STATUS_BUSY)
            self.open_delay_timer = CallbackTimer(open_delay, self._open_shutter)
            self.open_delay_timer.start()
        else:
            self._open_shutter()

    def _close_shutter(self):
        """
        Privater method, which must be overloaded.
        Put code, which open the shutter based on Device specification here.
        """
        raise NotImplementedError()

    @expose_method()
    def close_shutter(self, clear_auto_mode=True):
        """
        This is general method which close the shutter.
        It's calling _close_shutter method either directly or as a callback when CLOSE_DELAY is set.
        :param bool clear_auto_mode: Clear auto mode flag
        """
        self.clear_close_delay()
        self.set_value(ATTR_AUTO_MODE, not clear_auto_mode)

        close_delay = self.get_value(ATTR_CLOSE_DELAY)
        if close_delay:
            self.set_status(STATUS_BUSY)
            self.close_delay_timer = CallbackTimer(close_delay, self._close_shutter)
            self.close_delay_timer.start()
        else:
            self._close_shutter()

    def is_opened(self) -> bool:
        """
        Return True if Shutter is open, False otherwise.
        :return bool: Open state
        """
        raise NotImplementedError()

    def clear_open_delay(self):
        """
        Stop / clear open delay timer
        """
        if self.open_delay_timer is not None:
            self.open_delay_timer.stop()

    def clear_close_delay(self):
        """
        Stop / clear close delay timer
        """
        if self.close_delay_timer is not None:
            self.close_delay_timer.stop()

    @expose_method()
    def stop(self):
        self.clear_open_delay()
        self.clear_close_delay()
        self.set_status(STATUS_IDLE)

    def close(self) -> bool:
        self.clear_open_delay()
        self.clear_close_delay()
        return Device.close(self)


class DeviceShutterAttribute(DeviceShutter):
    """
    This is Shutter implemented by using Device and its Attribute.
    To use it, one must specify Device, Attribute and value for open and close state.
    Example of config file:

    Shutter_NI: &Shutter_NI !Device:kamzik3.devices.general.deviceShutter.DeviceShutterAttribute
        device: *NiCard
        device_id: Shutter
        attribute: [Digital outputs, port0/line0]
        opened_value: 1
        closed_value: 0
    """

    def __init__(
        self, device, attribute, opened_value, closed_value, device_id=None, config=None
    ):
        """
        :param Device device: Input device
        :param Union[list, tuple, str] attribute: Attribute to check
        :param mixed opened_value: value indicating open shutter
        :param mixed closed_value: value indicating close shutter
        :param str device_id: Device ID
        :param Union(dict, None) config: Device configuration
        """
        self.device = device
        self.attribute = Attribute.list_attribute(attribute)
        self.opened_value = opened_value
        self.closed_value = closed_value
        self.configured = False
        super().__init__(device_id, config)

    def _init_attributes(self):
        """
        Attributes;
            ATTR_OPENED_VALUE: Value indicating, that Shutter is Open
            ATTR_CLOSED_VALUE: Value indicating, that Shutter is Close
        """
        super()._init_attributes()
        self.create_attribute(
            ATTR_OPENED_VALUE, default_value=self.opened_value, default_type=np.float64
        )
        self.create_attribute(
            ATTR_CLOSED_VALUE, default_value=self.closed_value, default_type=np.float64
        )

    def connect(self, *args):
        self.device.attach_attribute_callback(
            ATTR_STATUS, self.set_status, key_filter=VALUE
        )

    def set_status(self, status):
        super().set_status(status)
        if status in READY_DEVICE_STATUSES and not self.configured:
            self.handle_configuration()

    def handle_configuration(self):
        """
        When configuring, set UNIT of the opened and closed value to unit of the Attribute.
        :return:
        """
        self.configured = True
        opened_value = device_units(self.device, self.attribute, self.opened_value)
        closed_value = device_units(self.device, self.attribute, self.closed_value)
        # noinspection PyStringFormat
        self.set_attribute([ATTR_OPENED_VALUE, UNIT], "{:~}".format(opened_value.u))  # type: ignore
        # noinspection PyStringFormat
        self.set_attribute([ATTR_CLOSED_VALUE, UNIT], "{:~}".format(closed_value.u))  # type: ignore
        self.device.attach_attribute_callback(
            self.attribute, self.shutter_value_changed, key_filter=VALUE
        )

    def shutter_value_changed(self, value):
        """
        Attribute callback, connected to the input Device Attribute.
        Callback is called when Value of the Attribute changed.
        :param mixed value: new value
        """
        self.set_value(ATTR_OPENED, self.is_opened(value))

    def _open_shutter(self):
        """
        Open shutter logic.
        By setting Status to IDLE, we are clearing delay BUSY status.
        """
        self.clear_open_delay()
        self.set_status(STATUS_IDLE)
        self.device.set_value(self.attribute, self.get_value(ATTR_OPENED_VALUE))

    def _close_shutter(self):
        """
        Close shutter logic.
        By setting Status to IDLE, we are clearing delay BUSY status.
        """
        self.clear_close_delay()
        self.set_status(STATUS_IDLE)
        self.device.set_value(self.attribute, self.get_value(ATTR_CLOSED_VALUE))

    # Note: I don't know if disabling arguments-differ is correct here. Might be a legitimate warning...
    # pylint: disable=arguments-differ
    def is_opened(self, current_value=None) -> bool:
        """
        Check if Shutter is open or close.
        :param mixed current_value: Current value of Device Attribute
        :return bool: Open state
        """
        opened_value = self.get_value(ATTR_OPENED_VALUE)
        if current_value is None:
            return False

        if self.device.get_attribute(self.attribute)[TYPE] == "bool":
            # Attribute is bool type
            opened = self.device.get_value(self.attribute) and opened_value
            self.set_value(ATTR_OPENED_AT, int(opened) * 100)
            return bool(self.device.get_value(self.attribute) and opened_value)

        current_value = self.device.get_value(self.attribute)
        tolerance = self.device.get_attribute(self.attribute + [TOLERANCE])

        try:
            self.set_value(
                ATTR_OPENED_AT, (float(current_value) / float(opened_value)) * 100
            )
        except ZeroDivisionError:
            self.set_value(ATTR_OPENED_AT, 100)
        return bool(
            (opened_value - tolerance[0])
            <= current_value
            <= (opened_value + tolerance[1])
        )

    @expose_method()
    def stop(self):
        DeviceShutter.stop(self)
        self.device.stop()

    def close(self) -> bool:
        """
        Close Device connection.
        """
        self.device.detach_attribute_callback(
            self.attribute, self.shutter_value_changed
        )
        return super().close()



import json
import logging
from threading import Thread

from kamzik3.devices.deviceClient import DeviceClient

import kamzik3
from kamzik3 import CommandFormatException, DeviceServerError, DeviceError
from kamzik3.constants import *
from kamzik3.snippets.snippetDataAccess import rgetattr
from kamzik3.snippets.snippetsJson import JsonKamzikHook, JsonKamzikEncoder
from kamzik3.snippets.snippetsZmq import Worker


class DeviceWorker(Worker, Thread):
    """
    This is Worker which servers the purpose of dealing with requests on Device.
    Server spawns multiple workers and then executing command from the client on free worker.
    This Worker is executing requests on the Device which is registered in the Session.
    DeviceClient sends request to the Server, then Server delegates this request to the Worker.
    Worker is then formulating response, which is send back to the client.
    """

    def __init__(self, id, master_host, master_port, zmq_context=None):
        """
        Initiate worker.
        Each worker, can respond to specific set of commands.
        Therefore we have to set instruction_set list.
        :param id:
        :param master_host:
        :param master_port:
        :param zmq_context:
        """
        self.logger = logging.getLogger("Console.DeviceWorker@{}".format(id))
        Thread.__init__(self)
        Worker.__init__(self, id, master_host, master_port, zmq_context)
        # Always override instruction set after worker initialization
        self.instructions_set += [
            INSTRUCTION_COMMAND,
            INSTRUCTION_GET,
            INSTRUCTION_SET,
            INSTRUCTION_METHOD,
            INSTRUCTION_POLL,
            INSTRUCTION_INIT,
        ]

    def get_payload(self, request):
        """
        Get input data, process and return result.
        :param list request: input data
        :return list:  [status, token, msg_type, response]
        """
        # Setting default response values
        status, token, response = RESPONSE_ERROR, 0, "Unknown error"

        try:
            instruction, device_id = request[0], request[1].decode()
            device = kamzik3.session.get_device(device_id)

            # Execute command directly on devices
            if instruction == INSTRUCTION_COMMAND:
                command, with_token = request[2].decode(), int(request[3])
                token = device.command(command, with_token=int(with_token))
                status, response = RESPONSE_OK, str(len(request[3]))

            # Get devices attribute
            elif instruction == INSTRUCTION_GET:
                attribute = json.loads(request[2].decode(), object_hook=JsonKamzikHook)
                response = device._get(attribute)
                response = json.dumps(
                    response, cls=JsonKamzikEncoder, ensure_ascii=True
                )
                status, token = RESPONSE_OK, 0

            # Set devices attribute
            elif instruction == INSTRUCTION_SET:
                attribute, attribute_value = json.loads(
                    request[2].decode(), object_hook=JsonKamzikHook
                )
                set_token = device.set_attribute(attribute, attribute_value)
                status, token, response = RESPONSE_OK, set_token, str(len(request[2]))

            # Execute devices method with attributes
            elif instruction == INSTRUCTION_METHOD:
                method, attributes = request[2].decode(), json.loads(
                    request[3].decode(), object_hook=JsonKamzikHook
                )
                response = rgetattr(device, method)(**attributes)
                response = json.dumps(
                    response, cls=JsonKamzikEncoder, ensure_ascii=True
                )
                status, token = RESPONSE_OK, 0

            # Poll devices for activity
            elif instruction == INSTRUCTION_POLL:
                status, token, response = RESPONSE_OK, 0, "1"

            # Init devices
            elif instruction == INSTRUCTION_INIT:
                if device is None or not device.in_statuses(READY_DEVICE_STATUSES):
                    raise DeviceServerError(
                        "Device {} is not registered on the server or publisher is not ready".format(
                            device_id
                        )
                    )
                else:
                    device_proxy = False
                    if isinstance(device, DeviceClient):
                        device_proxy = (device.host, device.port)
                    device_publisher = (
                        kamzik3.session.publisher.host,
                        kamzik3.session.publisher.port,
                    )
                    response = json.dumps(
                        (
                            device.attributes,
                            device.attributes_sharing_map,
                            device.exposed_methods,
                            device.qualified_name,
                            device_proxy,
                            device_publisher,
                        ),
                        cls=JsonKamzikEncoder,
                        ensure_ascii=True,
                    )
                    status, token = RESPONSE_OK, 0
            # None of above, request not implemented
            else:
                status, token, response = RESPONSE_ERROR, 0, str(len(request[3]))

        except CommandFormatException as e:
            status, token, response = RESPONSE_ERROR, 0, "Command format error"
            self.logger.error("Command format error {} {}".format(str(request), e))
        except DeviceServerError as e:
            status, token, response = RESPONSE_ERROR, 0, "Device server error"
            self.logger.error("Device server error {} {}".format(str(request), e))
        except DeviceError as e:
            status, token, response = RESPONSE_ERROR, 0, "Command error: {}".format(e)
            self.logger.error("Command error {} {}".format(str(request), e))
        except (AttributeError, KeyError) as e:
            status, token, response = RESPONSE_ERROR, 0, "Attribute error: {}".format(e)
            self.logger.error("Attribute error {} {}".format(str(request), e))
        except Exception as e:
            status, token, response = RESPONSE_ERROR, 0, "Unknown error: {}".format(e)
            self.logger.error("Unknown error {} {}".format(str(request), e))
        finally:
            # FIXME: Is it an error if an exception gets swallowed?
            return [status.encode(), str(token).encode(), MSG_JSON, response.encode()]



import re
import time

import bidict
import numpy as np
from kamzik3.snippets.snippetsDecorators import expose_method

import kamzik3
from kamzik3.constants import *
from kamzik3.devices.deviceSocket import DeviceSocket

VAL_NEGATIVE = "Negative"
VAL_POSITIVE = "Positive"
VAL_POWER_HOLD = "Power hold"
VAL_POWER_RELEASED = "Power released"
VAL_OFF = "Off"
VAL_UP_AND_DOWN = "Up and down"
VAL_UP_ONLY = "Up only"
VAL_UP_ONLY_SPECIAL = "Up only - special"
VAL_UP_ONLY_SPECIAL_2 = "Up only - special 2"


class DeviceProbusV(DeviceSocket):
    terminator = b"\n"
    push_commands_max = 1
    polarity_modes = bidict.frozenbidict(
        {
            VAL_POSITIVE: 0,
            VAL_NEGATIVE: 1,
        }
    )
    ramping_modes = bidict.frozenbidict(
        {
            VAL_OFF: 0,
            VAL_UP_AND_DOWN: 1,
            VAL_UP_ONLY: 2,
            VAL_UP_ONLY_SPECIAL: 3,
            VAL_UP_ONLY_SPECIAL_2: 4,
        }
    )
    error_check_pattern = re.compile(r"E(\d+)")
    error_description = {
        "E0": "No Error",
        "E1": "No data available",
        "E2": "Unknown register type",
        "E4": "Invalid argument",
        "E5": "Argument out of range",
        "E6": "Register is read only",
        "E7": "Receive overflow",
        "E8": "EEPROM is write protected",
        "E9": "Address error",
        "E10": "Unknown SCPI command",
        "E11": "Not allowed Trigger-on-Talk",
        "E12": "Invalid argument in ~Tn command",
        "E13": "Invalid N-value",
        "E14": "Register is write only",
        "E15": "String too long",
        "E16": "Wrong checksum",
    }

    # pylint: disable=unused-argument
    def __init__(
        self, host, port=2101, udp_host=None, udp_port=None, device_id=None, config=None
    ):
        DeviceSocket.__init__(self, host, port, device_id, config)

    def _init_attributes(self):
        DeviceSocket._init_attributes(self)

        self.create_attribute(
            ATTR_VOLTAGE_SETPOINT,
            default_type=np.float16,
            min_value=0,
            max_value=2000,
            decimals=3,
            unit="V",
            set_function=self.set_voltage,
        )
        self.create_attribute(
            ATTR_CURRENT_SETPOINT,
            default_type=np.float16,
            min_value=0,
            max_value=1.5,
            decimals=3,
            unit="A",
            set_function=self.set_current,
        )
        self.create_attribute(
            ATTR_VOLTAGE_READOUT,
            readonly=True,
            default_type=np.float16,
            min_value=0,
            decimals=3,
            unit="V",
        )
        self.create_attribute(
            ATTR_CURRENT_READOUT,
            readonly=True,
            default_type=np.float16,
            min_value=0,
            decimals=6,
            unit="A",
        )
        self.create_attribute(
            ATTR_OUTPUT_RELEASED,
            default_value=None,
            default_type=bool,
            set_function=self.set_release_output,
        )
        self.create_attribute(
            ATTR_DIGITAL_PROGRAMMING_ACTIVE,
            readonly=True,
            default_value=None,
            default_type=bool,
        )
        self.create_attribute(
            ATTR_ANALOGUE_PROGRAMMING_ACTIVE,
            readonly=True,
            default_value=None,
            default_type=bool,
        )
        self.create_attribute(
            ATTR_CONTROL_VOLTAGE, readonly=True, default_value=None, default_type=bool
        )
        self.create_attribute(
            ATTR_CONTROL_CURRENT, readonly=True, default_value=None, default_type=bool
        )
        self.create_attribute(
            ATTR_CALIBRATION_MODE, readonly=True, default_value=None, default_type=bool
        )
        self.create_attribute(
            ATTR_POLARITY_REVERSAL, readonly=True, default_value=None, default_type=bool
        )
        self.create_attribute(
            ATTR_CURRENT_ON_TARGET, readonly=True, default_value=None, default_type=bool
        )
        self.create_attribute(
            ATTR_VOLTAGE_ON_TARGET, readonly=True, default_value=None, default_type=bool
        )
        self.create_attribute(
            ATTR_RAMP_VOLTAGE_MODE,
            default_value=None,
            default_type=self.ramping_modes.keys(),
            set_function=self.set_voltage_ramping_mode,
        )
        self.create_attribute(
            ATTR_RAMP_CURRENT_MODE,
            default_value=None,
            default_type=self.ramping_modes.keys(),
            set_function=self.set_current_ramping_mode,
        )
        self.create_attribute(
            ATTR_OUTPUT_POLARITY,
            default_value=None,
            default_type=self.polarity_modes.keys(),
            set_function=self.set_polarity,
        )
        self.create_attribute(
            ATTR_VOLTAGE_RAMPING_RATE,
            default_type=np.float16,
            min_value=0,
            max_value=10000,
            decimals=3,
            unit="V/s",
            set_function=self.set_voltage_ramping_rate,
        )
        self.create_attribute(
            ATTR_CURRENT_RAMPING_RATE,
            default_type=np.float16,
            min_value=0,
            max_value=10000,
            decimals=3,
            unit="A/s",
            set_function=self.set_current_ramping_rate,
        )

    def handle_configuration(self):
        start_at = time.time()

        def _finish_configuration(*_, **__):
            self._config_commands()
            self._config_attributes()
            self.start_polling()
            self.set_status(STATUS_CONFIGURED)
            self.logger.info(
                "Device configuration took {} sec.".format(time.time() - start_at)
            )

        self.get_digital_programming_output_bit()
        self.get_analog_programming_active_bit()
        self.get_release_output_bit()
        self.get_control_voltage_bit()
        self.get_control_current_bit()
        self.get_polarity_reversal_bit()
        self.get_output_polarity_bit()
        self.get_voltage_ramping_rate()
        self.get_current_ramping_rate()
        self.get_voltage_ramping_mode()
        self.get_current_ramping_mode()
        # Set output measurement for V and A to normal precision
        self.command(">S0H 0\r\n")
        self.command(">S1H 0\r\n", callback=_finish_configuration)

    def handle_readout(self, readout_buffer):
        command, output, callback, token = DeviceSocket.handle_readout(
            self, readout_buffer
        )
        # Strip white characters from output
        error_code = self.error_check_pattern.match(output)
        if error_code is not None:
            if (int(error_code.group(1))) != 0:
                # We have error code
                self.handle_command_error(
                    command, self.error_description.get(output, "Unknown error")
                )
                return None
        else:
            (_request, value) = output.split(":")
            if command == ">S0?\r\n":
                self.set_raw_value(ATTR_VOLTAGE_SETPOINT, float(value))
            elif command == ">S1?\r\n":
                self.set_raw_value(ATTR_CURRENT_SETPOINT, float(value) * 1000)
            elif command == ">BON?\r\n":
                self.set_raw_value(ATTR_OUTPUT_RELEASED, bool(int(value)))
            elif command == ">KS?\r\n":
                self.set_raw_value(ATTR_CONTROL_CURRENT, bool(int(value[0])))
                self.set_raw_value(ATTR_CONTROL_VOLTAGE, bool(int(value[1])))
                self.set_raw_value(ATTR_OUTPUT_RELEASED, bool(int(value[2])))
                self.set_raw_value(ATTR_CALIBRATION_MODE, bool(int(value[5])))
                self.set_status(STATUS_BUSY if bool(int(value[2])) else STATUS_IDLE)
            elif command == ">DSD?\r\n":
                self.set_raw_value(ATTR_DIGITAL_PROGRAMMING_ACTIVE, bool(int(value)))
            elif command == ">DSA?\r\n":
                self.set_raw_value(ATTR_ANALOGUE_PROGRAMMING_ACTIVE, bool(int(value)))
            elif command == ">DVR?\r\n":
                self.set_raw_value(ATTR_CONTROL_VOLTAGE, bool(int(value)))
            elif command == ">DIR?\r\n":
                self.set_raw_value(ATTR_CONTROL_CURRENT, bool(int(value)))
            elif command == ">DCAL?\r\n":
                self.set_raw_value(ATTR_CALIBRATION_MODE, bool(int(value)))
            elif command == ">BX?\r\n":
                self.set_raw_value(ATTR_POLARITY_REVERSAL, bool(int(value)))
            elif command == ">DX?\r\n":
                self.set_raw_value(
                    ATTR_OUTPUT_POLARITY, self.polarity_modes.inv.get(int(value))
                )
            elif command == ">S0R?\r\n":
                self.set_raw_value(ATTR_VOLTAGE_RAMPING_RATE, float(value))
            elif command == ">S1R?\r\n":
                self.set_raw_value(ATTR_CURRENT_RAMPING_RATE, float(value))
            elif command == ">M0?\r\n":
                value = float(value)
                if self.get_value(ATTR_POLARITY_REVERSAL):
                    value *= -1
                self.set_raw_value(ATTR_VOLTAGE_READOUT, value)
            elif command == ">M1?\r\n":
                self.set_raw_value(ATTR_CURRENT_READOUT, float(value) * 1000)
            elif command == ">S0S?\r\n":
                self.set_raw_value(ATTR_VOLTAGE_ON_TARGET, bool(int(value)))
            elif command == ">S1S?\r\n":
                self.set_raw_value(ATTR_CURRENT_ON_TARGET, bool(int(value)))
            elif command == ">S0?\r\n":
                self.set_raw_value(ATTR_VOLTAGE_SETPOINT, float(value))
            elif command == ">S0B?\r\n":
                self.set_raw_value(
                    ATTR_RAMP_VOLTAGE_MODE, self.ramping_modes.inv.get(int(value))
                )
            elif command == ">S1B?\r\n":
                self.set_raw_value(
                    ATTR_RAMP_CURRENT_MODE, self.ramping_modes.inv.get(int(value))
                )
        if callback is not None:
            self.handle_readout_callback(callback, command, readout_buffer)
        if token:
            kamzik3.session.publisher.push_message(
                self.device_id, [command, readout_buffer], token
            )

        return command, output, callback, token

    def start_polling(self):
        DeviceSocket.start_polling(self)
        self.poll_command(">S0?\r\n", 1000)
        self.poll_command(">S1?\r\n", 1000)
        self.poll_command(">M0?\r\n", 500)
        self.poll_command(">M1?\r\n", 500)
        self.poll_command(">KS?\r\n", 300)
        self.poll_command(">S0S?\r\n", 1000)
        self.poll_command(">S1S?\r\n", 1000)

    def get_voltage_ramping_mode(self):
        self.command(">S0B?\r\n")

    def get_current_ramping_mode(self):
        self.command(">S1B?\r\n")

    def get_voltage(self):
        self.command(">S0?\r\n")

    def get_current(self):
        self.command(">S1?\r\n")

    def get_release_output_bit(self):
        self.command(">BON?\r\n")

    def get_polarity_reversal_bit(self):
        self.command(">BX?\r\n")

    def get_control_voltage_bit(self):
        self.command(">DVR?\r\n")

    def get_control_current_bit(self):
        self.command(">DIR?\r\n")

    def get_output_polarity_bit(self):
        self.command(">DX?\r\n")

    def get_digital_programming_output_bit(self):
        self.command(">DSD?\r\n")

    def get_analog_programming_active_bit(self):
        self.command(">DSA?\r\n")

    def get_voltage_readout(self):
        self.command(">M0?\r\n")

    def get_current_readout(self):
        self.command(">M1?\r\n")

    def get_offset(self):
        self.command(">CM0O?\r\n")
        self.command(">CM1O?\r\n")

    def get_gain(self):
        self.command(">CM0GP?\r\n")
        self.command(">CM0GN?\r\n")

    def get_voltage_ramping_rate(self):
        self.command(">S0R?\r\n")

    def get_current_ramping_rate(self):
        self.command(">S1R?\r\n")

    def get_state(self):
        self.command(">KS?\r\n")

    def set_voltage(self, value):
        self.command(f">S0 {value}\r\n")

    def set_current(self, value):
        value = float(value) * 1e-3
        self.command(f">S1 {value:.5e}\r\n")

    def set_release_output(self, value):
        self.command(f">BON {int(value)}\r\n")

    def set_polarity(self, value):
        self.command(f">BX {self.polarity_modes.get(value)}\r\n")

    def set_voltage_ramping_rate(self, value):
        self.command(f">S0R {value}\r\n")

    def set_current_ramping_rate(self, value):
        self.command(f">S1R {value}\r\n")

    def set_voltage_ramping_mode(self, value):
        self.command(f">S0B {self.ramping_modes.get(value)}\r\n")

    def set_current_ramping_mode(self, value):
        self.command(f">S1B {self.ramping_modes.get(value)}\r\n")

    @expose_method()
    def release_output(self):
        self.set_release_output(True)

    @expose_method()
    def stop(self):
        self.set_release_output(False)
